# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016, 2017, 2018
# --------------------------------------------------------------------------
# Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis

"""
This module handles the parameters that can be assigned to CP Optimizer to configure
the solving of a model.

The class `CpoParameters` contains the list of modifiable parameters
expressed as properties with getters and setters.
For the parameters that require special values, those values are given as constants.

Changing the value of a parameter can be done in multiple ways.
For example, the `TimeLimit` can be set to 60s :

   * `params.TimeLimit = 60`
   * `params.set_TimeLimit(60)`
   * `params['TimeLimit'] = 60`
   * `params.set_attribute('TimeLimit', 60)`

Retrieving the value of a parameter can be done in the same way using:

   * `v = params.TimeLimit`
   * `v = params.get_TimeLimit()`
   * `v = params['TimeLimit']`
   * `v = params.get_attribute('TimeLimit')`

If a parameter is not set, the value returned by the first two access forms is None.
The last access form (element of a dictionary) raises an exception.

Setting a parameter value to None is equivalent to force its default value.
This may be for example useful to reset at solve time a parameter that has been set at model level.

Getting the list of all parameters that have been changed can be done by calling the method `keys()`.

Note that the *PEP8* naming convention is not applied here, to keep parameter names as they
are in the solver, so that they can be referenced in solver logs.

Summary of parameters
---------------------

The following list gives the summary of all public parameters.

**Display and output**

 * :attr:`~CpoParameters.LogVerbosity`: Determines the verbosity of the search log.
   The value is a symbol in ['Quiet', 'Terse', 'Normal', 'Verbose']. Default value is 'Normal'.
 * :attr:`~CpoParameters.LogPeriod`: Controls how often the log information is displayed.
   The value is an integer strictly greater than 0. Default value is 1000.
 * :attr:`~CpoParameters.WarningLevel`: Level of warnings issued by CP Optimizer when a solve is launched.
   The value is an integer in [0..3]. Default value is 2.
 * :attr:`~CpoParameters.PrintModelDetailsInMessages`: Controls printing of additional information on error and warning messages.
   The value is a symbol in ['On', 'Off']. Default value is 'On'.
 * :attr:`~CpoParameters.ModelAnonymizer`: Controls anonymization of a model dumped via dumpModel.
   The value is a symbol in ['On', 'Off']. Default value is 'Off'.
 * :attr:`~CpoParameters.UseFileLocations`: Controls whether location information (file, line) is added to the model.
   The value is a symbol in ['On', 'Off']. Default value is 'On'.
 * :attr:`~CpoParameters.LogSearchTags`: Controls the log activation.
   The value is a symbol in ['On', 'Off']. Default value is 'Off'.
 * :attr:`~CpoParameters.KPIDisplay`: Controls the display of the KPI values in the log.
   The value is a symbol in ['SingleLine', 'MultipleLines']. Default value is 'SingleLine'.
   
**Presolve**

 * :attr:`~CpoParameters.Presolve`: Controls the presolve of the model to produce more compact formulations and to achieve more domain reduction.

**Optimality tolerances**

 * :attr:`~CpoParameters.OptimalityTolerance`: Absolute tolerance on the objective value for optimization models.
   The value is a positive float. Default value is 1e-09.
 * :attr:`~CpoParameters.RelativeOptimalityTolerance`: Relative tolerance on the objective value for optimization models.
   The value is a non-negative float. Default value is 0.0001.

**Search control**

 * :attr:`~CpoParameters.Workers`: Number of workers to run in parallel to solve the model.
   The value is a positive integer. Default value is Auto.
 * :attr:`~CpoParameters.SearchType`: Type of search that is applied when solving a problem.
   The value is a symbol in ['DepthFirst', 'Restart', 'MultiPoint', 'IterativeDiving', 'Auto']. Default value is 'Auto'.
 * :attr:`~CpoParameters.RandomSeed`: Seed of the random generator used by search strategies.
   The value is a non-negative integer. Default value is 0.
 * :attr:`~CpoParameters.RestartFailLimit`: Controls the number of failures that must occur before restarting search.
   The value is an integer greater than 0. Default value is 100.
 * :attr:`~CpoParameters.RestartGrowthFactor`: Controls the increase of the number of failures between restarts.
   The value is a float greater or equal to 1. Default value is 1.15.
 * :attr:`~CpoParameters.DynamicProbing`: Controls probing carried out during search.
   The value is a symbol in ['On', 'Off', 'Auto']. Default value is 'Auto'.
 * :attr:`~CpoParameters.DynamicProbingStrength`: Controls the effort which is dedicated to dynamic probing.
   The value is a float in [0.001..1000]. Default value is 0.03.
 * :attr:`~CpoParameters.MultiPointNumberOfSearchPoints`: Controls the number of solutions manipulated by the multi-point search algorithm.
   The value is an integer strictly greater than 1. Default value is 30.
 * :attr:`~CpoParameters.TemporalRelaxation`: Advanced parameter can be used to control the usage of a temporal relaxation.
   Possible values are 'On' or 'Off'.
 * :attr:`~CpoParameters.FailureDirectedSearch`: Controls usage of failure-directed search.
   Possible values are 'On' or 'Off'.
 * :attr:`~CpoParameters.FailureDirectedSearchEmphasis`: Controls how much time CP Optimizer invests into failure-directed search once it is started.
   The value is a non-negative integer, or None (default) that does not set any limit.
 * :attr:`~CpoParameters.FailureDirectedSearchMaxMemory`: Controls the maximum amount of memory available to failure-directed search
   The value is a non-negative integer, or None that does not set any limit. Default value is 104857600.
 * :attr:`~CpoParameters.AutomaticReplay`: Low-level control of the behavior of *solve()* and *next()*.
   Possible values are 'On' or 'Off'.

**Search limits**

 * :attr:`~CpoParameters.TimeLimit`: Limits the CPU or elapsed time spent solving before terminating a search.
   The value is a non-negative float, or None (default) that does not set any limit.
 * :attr:`~CpoParameters.TimeMode`: Defines how time is measured in CP Optimizer.
   The value is a symbol in ['CPUTime', 'ElapsedTime']. Default value is 'ElapsedTime'.
 * :attr:`~CpoParameters.FailLimit`: Limits the number of failures that can occur before terminating the search.
   The value is a non-negative integer, or None (default) that does not set any limit.
 * :attr:`~CpoParameters.ChoicePointLimit`: Limits the number of choice points that are created before terminating a search.
   The value is a non-negative integer, or None (default) that does not set any limit.
 * :attr:`~CpoParameters.BranchLimit`: Limits the number of branches that are made before terminating a search.
   The value is a non-negative integer, or None (default) that does not set any limit.
 * :attr:`~CpoParameters.SolutionLimit`: Limits the number of feasible solutions that are found before terminating a search.
   The value is a non-negative integer, or None (default) that does not set any limit.

**Inference levels for constraint propagation**

For all following attributes, possible values are 'Default', 'Low', 'Basic', Medium' or 'Extended'. Default value is 'Default'.

 * :attr:`~CpoParameters.DefaultInferenceLevel`: General inference level for constraints whose particular inference level is Default.
 * :attr:`~CpoParameters.AllDiffInferenceLevel`: Inference level for every constraint *allDiff*.
 * :attr:`~CpoParameters.DistributeInferenceLevel`: Inference level for every constraint *Distribute*.
 * :attr:`~CpoParameters.CountInferenceLevel`: Inference level for every constraint *Count* extracted to the invoked CP instance.
 * :attr:`~CpoParameters.CountDifferentInferenceLevel`: Inference level for every constraint *CountDifferent*.
 * :attr:`~CpoParameters.SequenceInferenceLevel`: Inference level for every constraint *Sequence*.
 * :attr:`~CpoParameters.AllMinDistanceInferenceLevel`: Inference level for every constraint *allMinDistance*.
 * :attr:`~CpoParameters.ElementInferenceLevel`: Inference level for every *element* constraint.
 * :attr:`~CpoParameters.PrecedenceInferenceLevel`: Inference level for precedence constraints between interval variables.
 * :attr:`~CpoParameters.IntervalSequenceInferenceLevel`: Inference level for the maintenance of the domain of every interval sequence variable.
 * :attr:`~CpoParameters.NoOverlapInferenceLevel`: Inference level for every constraint NoOverlap extracted.
 * :attr:`~CpoParameters.CumulFunctionInferenceLevel`: Inference level for constraints on expression *CumulFunctionExpr*.
 * :attr:`~CpoParameters.StateFunctionInferenceLevel`: Inference level for constraints on state functions *StateFunction*.

**Conflict refiner**

 * :attr:`~CpoParameters.ConflictRefinerTimeLimit`: Limits the CPU time spent before terminating the conflict refiner.
   The value is a non-negative float, or None (default) that does not set any limit.
 * :attr:`~CpoParameters.ConflictRefinerIterationLimit`: Limits the number of iterations that are made before terminating the conflict refiner.
   The value is a non-negative integer, or None (default) that does not set any limit.
 * :attr:`~CpoParameters.ConflictRefinerBranchLimit`: Limits the total number of branches that are made before terminating the conflict refiner.
   The value is a non-negative integer, or None (default) that does not set any limit.
 * :attr:`~CpoParameters.ConflictRefinerFailLimit`: Limits the total number of failures that can occur before terminating the conflict refiner.
   The value is a non-negative integer, or None (default) that does not set any limit.
 * :attr:`~CpoParameters.ConflictRefinerOnVariables`: Specifies whether the conflict refiner should refine variables domains.
   The value is a symbol in ['On', 'Off']. Default value is 'Off'.


Detailed description
--------------------
"""

from docplex.cp.utils import Context, is_int, is_number
import warnings


###############################################################################
## Public constants
###############################################################################

# Set of all public parameter names
PUBLIC_PARAMETER_NAMES = \
    {"AllDiffInferenceLevel", "AllMinDistanceInferenceLevel", "AutomaticReplay", "BranchLimit", "ChoicePointLimit",
     "ConflictRefinerBranchLimit", "ConflictRefinerFailLimit", "ConflictRefinerIterationLimit",
     "ConflictRefinerOnVariables", "ConflictRefinerTimeLimit","CountDifferentInferenceLevel", "CountInferenceLevel",
     "CumulFunctionInferenceLevel", "DefaultInferenceLevel", "DistributeInferenceLevel", "DynamicProbing",
     "DynamicProbingStrength", "ElementInferenceLevel", "FailLimit", "FailureDirectedSearch",
     "FailureDirectedSearchEmphasis", "FailureDirectedSearchMaxMemory", "IntervalSequenceInferenceLevel", "LogPeriod",
     "LogSearchTags", "LogVerbosity", "ModelAnonymizer", "MultiPointNumberOfSearchPoints", "NoOverlapInferenceLevel",
     "OptimalityTolerance", "PrecedenceInferenceLevel", "Presolve", "PrintModelDetailsInMessages", "RandomSeed",
     "RelativeOptimalityTolerance", "RestartFailLimit", "RestartGrowthFactor", "SearchType", "SequenceInferenceLevel",
     "SolutionLimit", "StateFunctionInferenceLevel", "TemporalRelaxation", "TimeLimit", "TimeMode", "UseFileLocations",
     "WarningLevel", "Workers", "KPIDisplay"}

# Set of all private but accepted parameter names
PRIVATE_PARAMETER_NAMES = {"ObjectiveLimit",}

# Set of all authorized parameter names
ALL_PARAMETER_NAMES = PUBLIC_PARAMETER_NAMES | PRIVATE_PARAMETER_NAMES

# Symbolic parameter values
VALUE_AUTO             = 'Auto'
VALUE_OFF              = 'Off'
VALUE_ON               = 'On'
VALUE_DEFAULT          = 'Default'
VALUE_LOW              = 'Low'
VALUE_BASIC            = 'Basic'
VALUE_MEDIUM           = 'Medium'
VALUE_EXTENDED         = 'Extended'
VALUE_QUIET            = 'Quiet'
VALUE_TERSE            = 'Terse'
VALUE_NORMAL           = 'Normal'
VALUE_VERBOSE          = 'Verbose'
VALUE_DEPTH_FIRST      = 'DepthFirst'
VALUE_RESTART          = 'Restart'
VALUE_MULTI_POINT      = 'MultiPoint'
VALUE_ITERATIVE_DIVING = 'IterativeDiving'
VALUE_DIVERSE          = 'Diverse'
VALUE_CPU_TIME         = 'CPUTime'
VALUE_ELAPSED_TIME     = 'ElapsedTime'
VALUE_SINGLE_LINE      = 'SingleLine'
VALUE_MULTIPLE_LINES   = 'MultipleLines'

# Authorized sets of values
_INFERENCE_LEVELS = (VALUE_DEFAULT, VALUE_LOW, VALUE_BASIC, VALUE_MEDIUM, VALUE_EXTENDED)
_ON_OFF_AUTO = (VALUE_ON, VALUE_OFF, VALUE_AUTO)
_ON_OFF = (VALUE_ON, VALUE_OFF)
_KPI_DISPLAY = (VALUE_SINGLE_LINE, VALUE_MULTIPLE_LINES)
_SEARCH_TYPES = (VALUE_DEPTH_FIRST, VALUE_RESTART, VALUE_MULTI_POINT, VALUE_ITERATIVE_DIVING, VALUE_AUTO)


###############################################################################
## Public classes
###############################################################################

class CpoParameters(Context):
    """ Class for handling solving parameters
    """
    def __init__(self, **kwargs):
        """ Creates a new set of solving parameters.

        This constructor takes a variable number of optional arguments that allow to set parameters directly.
        For example:
        ::
           myparams = CpoParameters(TimeLimit=20, LogPeriod=5000))

        Args:
            kwargs: (Optional) Any individual parameter as defined in this class.
        """
        super(CpoParameters, self).__init__()
        for k, v in kwargs.items():
            self.__setattr__(k, v)


    def __setattr__(self, name, value):
        """ Set a parameter.
        This method calls appropriate setter if exists (property does not work in this context)
        to check the parameter value.
        Args:
            name:  Parameter name
            value: Parameter value
        """
        m = getattr(self, 'set_' + name, None)
        if m and callable(m):
            m(value)
        else:
            self.set_attribute(name, value)


    def reset_to_default(self):
        """ Reset all the parameters to their default value.

        Parameters are reset to their default value by being removed from this object.
        """
        self.clear()


    def add(self, ctx):
        """ Add another parameters to this one.

        All attributes of given parameters are set in this one, except if there value is None.
        If one value is another context, it is cloned before being set.

        Args:
            ctx:  Other context to add to this one.
        """
        for k, v in ctx.items():
            if v is not None:
                if isinstance(v, Context):
                    v = v.clone()
                self.set_attribute(k, v)


    def get_AllDiffInferenceLevel(self):
        """
        This parameter specifies the inference level for every constraint AllDiff extracted to the invoking
        CP instance. Possible values for this parameter are Default, Low, Basic, Medium, and Extended. The
        default value is Default, which allows the inference strength of all AllDiff constraints to be
        controlled via DefaultInferenceLevel.

        The value is a symbol in ['Default', 'Low', 'Basic', 'Medium', 'Extended']. Default value is 'Default'.
        """
        return self.get_attribute('AllDiffInferenceLevel')

    def set_AllDiffInferenceLevel(self, val):
        self._set_value_enum('AllDiffInferenceLevel', val, _INFERENCE_LEVELS)

    AllDiffInferenceLevel = property(get_AllDiffInferenceLevel, set_AllDiffInferenceLevel)


    def get_AllMinDistanceInferenceLevel(self):
        """
        This parameter specifies the inference level for every constraint AllMinDistance extracted to the
        invoked CP instance. Possible values for this parameter are Default, Low, Basic, Medium, and
        Extended. The default value is Default, which allows the inference strength of all AllMinDistance
        constraints to be controlled via DefaultInferenceLevel.

        The value is a symbol in ['Default', 'Low', 'Basic', 'Medium', 'Extended']. Default value is 'Default'.
        """
        return self.get_attribute('AllMinDistanceInferenceLevel')

    def set_AllMinDistanceInferenceLevel(self, val):
        self._set_value_enum('AllMinDistanceInferenceLevel', val, _INFERENCE_LEVELS)

    AllMinDistanceInferenceLevel = property(get_AllMinDistanceInferenceLevel, set_AllMinDistanceInferenceLevel)


    def get_AutomaticReplay(self):
        """
        This parameter is an advanced, low-level one for controlling the behavior of solve() and next().
        When the model being solved has an objective and solve is used, or when startNewSearch and next are
        used to produce multiple solutions, the solver may have a need to replay the last (or best) solution
        found. This can, in some cases, involve re-invoking the strategy which produced the solution.
        Normally this is only necessary if you use low level "Ilc" interfaces to specify problem elements
        not in the model (instance of Model). This parameter can take the values On or Off. The default
        value is On. A typical reason for setting this parameter to Off is, for instance, if you use your
        own custom goal (instance of IlcGoal), and this goal is not deterministic (does not do the same
        thing when executed twice). In this instance, the replay will not work correctly, and you can use
        this parameter to disable replay.

        This parameter is deprecated since release 2.3.

        The value is a symbol in ['On', 'Off']. Default value is 'On'.
        """
        warnings.warn("Parameter 'AutomaticReplay' is deprecated since release 2.3.", DeprecationWarning)
        return self.get_attribute('AutomaticReplay')

    def set_AutomaticReplay(self, val):
        warnings.warn("Parameter 'AutomaticReplay' is deprecated since release 2.3.", DeprecationWarning)
        self._set_value_enum('AutomaticReplay', val, _ON_OFF)

    AutomaticReplay = property(get_AutomaticReplay, set_AutomaticReplay)


    def get_BranchLimit(self):
        """
        This parameter limits the number of branches that are made before terminating a search. A branch is
        a decision made at a choice point in the search, a typical node being made up of two branches, for
        example: x == value and x != value. A branch is only counted at the moment a decision is executed,
        not when the two branches of the choice point are decided. A branch is counted even if the decision
        leads to an inconsistency (failure).

        The value is a non-negative integer, or None (default) that does not set any limit.
        """
        return self.get_attribute('BranchLimit')

    def set_BranchLimit(self, val):
        self._set_value_integer('BranchLimit', val)

    BranchLimit = property(get_BranchLimit, set_BranchLimit)


    def get_ChoicePointLimit(self):
        """
        This parameter limits the number of choice points that are created before terminating a search.

        The value is a non-negative integer, or None (default) that does not set any limit.
        """
        return self.get_attribute('ChoicePointLimit')

    def set_ChoicePointLimit(self, val):
        self._set_value_integer('ChoicePointLimit', val)

    ChoicePointLimit = property(get_ChoicePointLimit, set_ChoicePointLimit)


    def get_ConflictRefinerBranchLimit(self):
        """
        This parameter limits the total number of branches that are made before terminating the conflict
        refiner.

        The value is a non-negative integer, or None (default) that does not set any limit.
        """
        return self.get_attribute('ConflictRefinerBranchLimit')

    def set_ConflictRefinerBranchLimit(self, val):
        self._set_value_integer('ConflictRefinerBranchLimit', val)

    ConflictRefinerBranchLimit = property(get_ConflictRefinerBranchLimit, set_ConflictRefinerBranchLimit)


    def get_ConflictRefinerFailLimit(self):
        """
        This parameter limits the total number of failures that can occur before terminating the conflict
        refiner.

        The value is a non-negative integer, or None (default) that does not set any limit.
        """
        return self.get_attribute('ConflictRefinerFailLimit')

    def set_ConflictRefinerFailLimit(self, val):
        self._set_value_integer('ConflictRefinerFailLimit', val)

    ConflictRefinerFailLimit = property(get_ConflictRefinerFailLimit, set_ConflictRefinerFailLimit)


    def get_ConflictRefinerIterationLimit(self):
        """
        This parameter limits the number of iterations that are made before terminating the conflict
        refiner.

        The value is a non-negative integer, or None (default) that does not set any limit.
        """
        return self.get_attribute('ConflictRefinerIterationLimit')

    def set_ConflictRefinerIterationLimit(self, val):
        self._set_value_integer('ConflictRefinerIterationLimit', val)

    ConflictRefinerIterationLimit = property(get_ConflictRefinerIterationLimit, set_ConflictRefinerIterationLimit)


    def get_ConflictRefinerOnVariables(self):
        """
        This parameter specifies whether the conflict refiner should refine variables domains. Possible
        values for this parameter are On (conflict refiner will refine both constraints and variables
        domains) and Off (conflict refiner will only refine constraints).

        The value is a symbol in ['On', 'Off']. Default value is 'Off'.
        """
        return self.get_attribute('ConflictRefinerOnVariables')

    def set_ConflictRefinerOnVariables(self, val):
        self._set_value_enum('ConflictRefinerOnVariables', val, _ON_OFF)

    ConflictRefinerOnVariables = property(get_ConflictRefinerOnVariables, set_ConflictRefinerOnVariables)


    def get_ConflictRefinerTimeLimit(self):
        """
        This parameter limits the CPU time spent before terminating the conflict refiner.

        The value is a non-negative float, or None (default) that does not set any limit.
        """
        return self.get_attribute('ConflictRefinerTimeLimit')

    def set_ConflictRefinerTimeLimit(self, val):
        self._set_value_float('ConflictRefinerTimeLimit', val)

    ConflictRefinerTimeLimit = property(get_ConflictRefinerTimeLimit, set_ConflictRefinerTimeLimit)


    def get_CountDifferentInferenceLevel(self):
        """
        This parameter specifies the inference level for every constraint CountDifferent extracted to the
        invoking CP instance. Possible values for this parameter are Default, Low, Basic, Medium, and
        Extended. The default value is Default, which allows the inference strength of all CountDifferent
        constraints to be controlled via DefaultInferenceLevel.

        The value is a symbol in ['Default', 'Low', 'Basic', 'Medium', 'Extended']. Default value is 'Default'.
        """
        return self.get_attribute('CountDifferentInferenceLevel')

    def set_CountDifferentInferenceLevel(self, val):
        self._set_value_enum('CountDifferentInferenceLevel', val, _INFERENCE_LEVELS)

    CountDifferentInferenceLevel = property(get_CountDifferentInferenceLevel, set_CountDifferentInferenceLevel)


    def get_CountInferenceLevel(self):
        """
        This parameter specifies the inference level for every constraint Count extracted to the invoked CP
        instance. Possible values for this parameter are Default, Low, Basic, Medium, and Extended. The
        default value is Default, which allows the inference strength of all Count constraints to be
        controlled via DefaultInferenceLevel.

        The value is a symbol in ['Default', 'Low', 'Basic', 'Medium', 'Extended']. Default value is 'Default'.
        """
        return self.get_attribute('CountInferenceLevel')

    def set_CountInferenceLevel(self, val):
        self._set_value_enum('CountInferenceLevel', val, _INFERENCE_LEVELS)

    CountInferenceLevel = property(get_CountInferenceLevel, set_CountInferenceLevel)


    def get_CumulFunctionInferenceLevel(self):
        """
        This parameter specifies the inference level for constraints on expressions CumulFunctionExpr
        extracted to the invoked CP instance. Possible values for this parameter are Default, Low, Basic,
        Medium, and Extended. The default value is Default, which allows the inference strength of all
        constraints on CumulFunctionExpr to be controlled via DefaultInferenceLevel.

        The value is a symbol in ['Default', 'Low', 'Basic', 'Medium', 'Extended']. Default value is 'Default'.
        """
        return self.get_attribute('CumulFunctionInferenceLevel')

    def set_CumulFunctionInferenceLevel(self, val):
        self._set_value_enum('CumulFunctionInferenceLevel', val, _INFERENCE_LEVELS)

    CumulFunctionInferenceLevel = property(get_CumulFunctionInferenceLevel, set_CumulFunctionInferenceLevel)


    def get_DefaultInferenceLevel(self):
        """
        This parameter specifies the general inference level for constraints whose particular inference
        level is Default. Possible values for this parameter (in increasing order of inference strength) are
        Low, Basic, Medium, and Extended.

       The value is a symbol in ['Default', 'Low', 'Basic', 'Medium', 'Extended']. Default value is 'Default'.
         """
        return self.get_attribute('DefaultInferenceLevel')

    def set_DefaultInferenceLevel(self, val):
        self._set_value_enum('DefaultInferenceLevel', val, _INFERENCE_LEVELS)

    DefaultInferenceLevel = property(get_DefaultInferenceLevel, set_DefaultInferenceLevel)


    def get_DistributeInferenceLevel(self):
        """
        This parameter specifies the inference level for every constraint Distribute extracted to the
        invoked CP instance. Possible values for this parameter are Default, Low, Basic, Medium, and
        Extended. The default value is Default, which allows the inference strength of all Distribute
        constraints to be controlled via DefaultInferenceLevel.

        The value is a symbol in ['Default', 'Low', 'Basic', 'Medium', 'Extended']. Default value is 'Default'.
        """
        return self.get_attribute('DistributeInferenceLevel')

    def set_DistributeInferenceLevel(self, val):
        self._set_value_enum('DistributeInferenceLevel', val, _INFERENCE_LEVELS)

    DistributeInferenceLevel = property(get_DistributeInferenceLevel, set_DistributeInferenceLevel)


    def get_DynamicProbing(self):
        """
        This parameter controls probing carried out during search. Probing can be useful on some problems as
        it can make stronger inferences on combinations of constraints. Possible values for this parameter
        are On (dynamic probing is activated with a constant strength), Auto (dynamic probing is activated
        and its strength is adjusted adaptively) and Off (dynamic probing is deactivated). The strength of
        probing can be defined by parameter DynamicProbingStrength. Dynamic probing only has an effect when
        using the "Restart" (Restart) search type, on problems without interval variables.

        The value is a symbol in ['On', 'Off', 'Auto']. Default value is 'Auto'.
        """
        return self.get_attribute('DynamicProbing')

    def set_DynamicProbing(self, val):
        self._set_value_enum('DynamicProbing', val, _ON_OFF_AUTO)

    DynamicProbing = property(get_DynamicProbing, set_DynamicProbing)


    def get_DynamicProbingStrength(self):
        """
        This parameter controls the effort which is dedicated to dynamic probing. It is expressed as a
        factor of the total search effort: changing this parameter has no effect unless the DynamicProbing
        parameter is set to Auto or On. When DynamicProbing has value On, the probing strength is held
        constant throughout the search process. When DynamicProbing has value Auto, the probing strength
        starts off at the specified value and is thereafter adjusted automatically. Possible values for this
        parameter range from 0.001 to 1000. A value of 1.0 indicates that dynamic probing will consume a
        roughly equal amount of effort as the rest of the search. The default value of this parameter is
        0.03, meaning that around 3% of total search time is dedicated to dynamic probing.

        The value is a float in [0.001..1000]. Default value is 0.03.
        """
        return self.get_attribute('DynamicProbingStrength')

    def set_DynamicProbingStrength(self, val):
        self._set_value_float('DynamicProbingStrength', val, 0.001, 1000)

    DynamicProbingStrength = property(get_DynamicProbingStrength, set_DynamicProbingStrength)


    def get_ElementInferenceLevel(self):
        """
        This parameter specifies the inference level for every element constraint extracted to the invoked
        CP instance. Possible values for this parameter are Default, Low, Basic, Medium, and Extended. The
        default value is Default, which allows the inference strength of all element constraints to be
        controlled via DefaultInferenceLevel.

        The value is a symbol in ['Default', 'Low', 'Basic', 'Medium', 'Extended']. Default value is 'Default'.
        """
        return self.get_attribute('ElementInferenceLevel')

    def set_ElementInferenceLevel(self, val):
        self._set_value_enum('ElementInferenceLevel', val, _INFERENCE_LEVELS)

    ElementInferenceLevel = property(get_ElementInferenceLevel, set_ElementInferenceLevel)


    def get_FailLimit(self):
        """
        This parameter limits the number of failures that can occur before terminating the search.

        The value is a non-negative integer, or None (default) that does not set any limit.
        """
        return self.get_attribute('FailLimit')

    def set_FailLimit(self, val):
        self._set_value_integer('FailLimit', val)

    FailLimit = property(get_FailLimit, set_FailLimit)


    def get_FailureDirectedSearch(self):
        """
        This parameter controls usage of failure-directed search. Failure-directed search assumes that there
        is no (better) solution or that such a solution is very hard to find. Therefore it focuses on a
        systematic exploration of search space, first eliminating assignments that are most likely to fail.
        Failure-directed search is used only for scheduling problems (i.e. models containing interval
        variables) and only when the parameter SearchType is set to Restart or Auto. Legal values for the
        FailureDirectedSearch parameter are On (the default) and Off. When the value is On then CP Optimizer
        starts failure-directed search when other search strategies are (no longer) successful and when the
        memory necessary for the search does not exceed the value set by the FailureDirectedSearchMaxMemory
        parameter.

        The value is a symbol in ['On', 'Off']. Default value is 'On'.
        """
        return self.get_attribute('FailureDirectedSearch')

    def set_FailureDirectedSearch(self, val):
        self._set_value_enum('FailureDirectedSearch', val, _ON_OFF)

    FailureDirectedSearch = property(get_FailureDirectedSearch, set_FailureDirectedSearch)


    def get_FailureDirectedSearchEmphasis(self):
        """
        This parameter controls how much time CP Optimizer invests into failure-directed search once it is
        started. The default value Auto means that CP Optimizer observes the actual performance of
        failure-directed search and decides automatically how much time is invested. Any other value means
        that once failure-directed search has started, it is used by given number of workers. The value does
        not have to be integer. For example, value 1.5 means that first worker spends 100% of the time by
        failure-directed search, second worker 50% and remaining workers 0%. See also Workers For more
        information about failure-directed search see parameter FailureDirectedSearch.
        
        The value is a non-negative integer, or None (default) that does not set any limit.
        """
        return self.get_attribute('FailureDirectedSearchEmphasis')

    def set_FailureDirectedSearchEmphasis(self, val):
        self._set_value_float('FailureDirectedSearchEmphasis', val)

    FailureDirectedSearchEmphasis = property(get_FailureDirectedSearchEmphasis, set_FailureDirectedSearchEmphasis)


    def get_FailureDirectedSearchMaxMemory(self):
        """
        This parameter controls the maximum amount of memory (in bytes) available to failure-directed search
        (see FailureDirectedSearchMaxMemory). The default value is 104,857,600 (100MB). Failure-directed
        search can sometimes consume a lot of memory, especially when end times of interval variables are
        not bounded. Therefore it is usually not started immediately, but only when the effective horizon
        (time period over which CP Optimizer must reason) becomes small enough for failure-directed search
        to operate inside the memory limit specified by this parameter. For many types of scheduling
        problems, the effective horizon tends to reduce when CP Optimizer finds a better solution (often
        most significantly when the initial solution is found). Therefore, when each new solution is found,
        CP Optimizer decides whether or not to turn on failure-directed search. Note that this parameter
        does not influence the effectiveness of failure-directed search, once started. Its purpose is only
        to control the point at which failure-directed search will begin to function.

        The value is a non-negative integer, or None that does not set any limit. Default value is 104857600.
        """
        return self.get_attribute('FailureDirectedSearchMaxMemory')

    def set_FailureDirectedSearchMaxMemory(self, val):
        self._set_value_integer('FailureDirectedSearchMaxMemory', val)

    FailureDirectedSearchMaxMemory = property(get_FailureDirectedSearchMaxMemory, set_FailureDirectedSearchMaxMemory)


    def get_IntervalSequenceInferenceLevel(self):
        """
        This parameter specifies the inference level for the maintenance of the domain of every interval
        sequence variable IntervalSequenceVar extracted to the invoking CP instance. Possible values for
        this parameter are Default, Low, Basic, Medium, and Extended. The default value is Default, which
        allows the inference strength of all IntervalSequenceVar to be controlled via DefaultInferenceLevel.

        The value is a symbol in ['Default', 'Low', 'Basic', 'Medium', 'Extended']. default value is 'Default'.
        """
        return self.get_attribute('IntervalSequenceInferenceLevel')

    def set_IntervalSequenceInferenceLevel(self, val):
        self._set_value_enum('IntervalSequenceInferenceLevel', val, _INFERENCE_LEVELS)

    IntervalSequenceInferenceLevel = property(get_IntervalSequenceInferenceLevel, set_IntervalSequenceInferenceLevel)


    def get_KPIDisplay(self):
        """
        This parameter determines how KPIs are displayed in the log during the search.

        The value is a symbol in ['SingleLine', 'MultipleLines']. Default value is 'SingleLine'.

        *New in version 2.8, for CPO solver version 12.9.0 and higher.*
        """
        return self.get_attribute('KPIDisplay')

    def set_KPIDisplay(self, val):
        self._set_value_enum('KPIDisplay', val, _KPI_DISPLAY)

    KPIDisplay = property(get_KPIDisplay, set_KPIDisplay)


    def get_LogPeriod(self):
        """
        The CP Optimizer search log includes information that is displayed periodically. This parameter
        controls how often that information is displayed. By setting this parameter to a value of k, the log
        is displayed every k branches (search decisions).

        The value is an integer strictly greater than 0. Default value is 1000.
        """
        return self.get_attribute('LogPeriod')

    def set_LogPeriod(self, val):
        self._set_value_integer('LogPeriod', val, min=1)

    LogPeriod = property(get_LogPeriod, set_LogPeriod)


    # def get_LogSearchTags(self):
    #     """
    #     This parameter controls the log activation. When set to On, the engine will display failure tags
    #     (indices) in the engine log when solving the model. To specify the failures to explain, the member
    #     functions explainFailure(Int failureTag) or explainFailure(IntArray tagArray) needs to be called
    #     with the failure tags as the parameter. Several failures tags can be added. The member function
    #     clearExplanations() is used to clear the set of failure tags to be explained. To be able to see
    #     failure tags and explanations, the parameter SearchType must be set to DepthFirst and the parameter
    #     Workers to 1.
    #
    #     The value is a symbol in ['On', 'Off']. Default value is 'Off'.
    #     """
    #     return self.get_attribute('LogSearchTags')
    #
    # def set_LogSearchTags(self, val):
    #     self._set_value_enum('LogSearchTags', val, _ON_OFF)
    #
    # LogSearchTags = property(get_LogSearchTags, set_LogSearchTags)


    def get_LogVerbosity(self):
        """
        This parameter determines the verbosity of the search log. The possible values are Quiet, Terse,
        Normal, and Verbose. Mode Quiet does not display any information, the other modes display
        progressively more information. The default value is Normal. The CP Optimizer search log is meant
        for visual inspection only, not for mechanized parsing. In particular, the log may change from
        version to version of CP Optimizer in order to improve the quality of information displayed in the
        log. Any code based on the log output for correct functioning may have to be updated when a new
        version of CP Optimizer is released.

        The value is a symbol in ['Quiet', 'Terse', 'Normal', 'Verbose']. Default value is 'Normal'.
        """
        return self.get_attribute('LogVerbosity')

    def set_LogVerbosity(self, val):
        self._set_value_enum('LogVerbosity', val, (VALUE_QUIET, VALUE_TERSE, VALUE_NORMAL, VALUE_VERBOSE))

    LogVerbosity = property(get_LogVerbosity, set_LogVerbosity)


    def get_ModelAnonymizer(self):
        """
        This parameter controls anonymization of a model dumped in CPO file format. The legal values of this
        parameter are Off and On. The default is Off. When the anonymizer is off, then names of variables
        and constraints in the model may be found in the output file. When the anonymizer is on, names given
        to variables or constraints in the model will not be reflected in the output file and standard
        anonymized names will be used.

        The value is a symbol in ['On', 'Off']. Default value is 'Off'.
        """
        return self.get_attribute('ModelAnonymizer')

    def set_ModelAnonymizer(self, val):
        self._set_value_enum('ModelAnonymizer', val, _ON_OFF)

    ModelAnonymizer = property(get_ModelAnonymizer, set_ModelAnonymizer)


    def get_MultiPointNumberOfSearchPoints(self):
        """
        This parameter controls the number of (possibly partial) solutions manipulated by the multi-point
        search algorithm. The default value is 30. A larger value will diversify the search, with possible
        improvement in solution quality at the expense of a longer run time. A smaller value will intensify
        the search, resulting in faster convergence at the expense of solution quality. Note that memory
        consumption increases proportionally to this parameter, for each search point must store each
        decision variable domain.

        The value is an integer strictly greater than 1. Default value is 30.
        """
        return self.get_attribute('MultiPointNumberOfSearchPoints')

    def set_MultiPointNumberOfSearchPoints(self, val):
        self._set_value_integer('MultiPointNumberOfSearchPoints', val, min=2)

    MultiPointNumberOfSearchPoints = property(get_MultiPointNumberOfSearchPoints, set_MultiPointNumberOfSearchPoints)


    def get_NoOverlapInferenceLevel(self):
        """
        This parameter specifies the inference level for every constraint NoOverlap extracted to the invoked
        CP instance. Possible values for this parameter are Default, Low, Basic, Medium, and Extended. The
        default value is Default, which allows the inference strength of all NoOverlap constraints to be
        controlled via DefaultInferenceLevel.

        The value is a symbol in ['Default', 'Low', 'Basic', 'Medium', 'Extended']. Default value is 'Default'.
        """
        return self.get_attribute('NoOverlapInferenceLevel')

    def set_NoOverlapInferenceLevel(self, val):
        self._set_value_enum('NoOverlapInferenceLevel', val, _INFERENCE_LEVELS)

    NoOverlapInferenceLevel = property(get_NoOverlapInferenceLevel, set_NoOverlapInferenceLevel)


    def get_OptimalityTolerance(self):
        """
        This parameter sets an absolute tolerance on the objective value for optimization models. This means
        that when CP Optimizer reports an optimal solution found, then there is no solution which improves
        the objective by more than the value of this parameter.
        This parameter is used in conjunction with RelativeOptimalityTolerance. The optimality of a solution
        is proven if either of the two parameters' criteria is fulfilled.

        The value is a positive float. Default value is 1e-09.
        """
        return self.get_attribute('OptimalityTolerance', 1e-09)

    def set_OptimalityTolerance(self, val):
        self._set_value_float('OptimalityTolerance', val)

    OptimalityTolerance = property(get_OptimalityTolerance, set_OptimalityTolerance)


    def get_PrecedenceInferenceLevel(self):
        """
        This parameter specifies the inference level for precedence constraints between interval variables
        extracted to the invoking CP instance. Possible values for this parameter are Default, Low, Basic,
        Medium, and Extended. The default value is Default, which allows the inference strength for
        precedence constraints between interval variables to be controlled via DefaultInferenceLevel.

        The value is a symbol in ['Default', 'Low', 'Basic', 'Medium', 'Extended']. Default value is 'Default'.
        """
        return self.get_attribute('PrecedenceInferenceLevel')

    def set_PrecedenceInferenceLevel(self, val):
        self._set_value_enum('PrecedenceInferenceLevel', val, _INFERENCE_LEVELS)

    PrecedenceInferenceLevel = property(get_PrecedenceInferenceLevel, set_PrecedenceInferenceLevel)


    def get_Presolve(self):
        """
        This parameter controls the presolve of the model to produce more compact formulations and to
        achieve more domain reduction. Possible values for this parameter are On (presolve is activated) and
        Off (presolve is deactivated).

        The value is a symbol in ['On', 'Off']. Default value is 'On'.
        """
        return self.get_attribute('Presolve')

    def set_Presolve(self, val):
        self._set_value_enum('Presolve', val, _ON_OFF)

    Presolve = property(get_Presolve, set_Presolve)


    def get_PrintModelDetailsInMessages(self):
        """
        Whenever CP Optimizer prints an error or warning message, it can also print concerning part of the
        input model (in cpo file format). This parameter controls printing of this additional information.
        Possible values are On and Off, the Default value is 'On'. See also WarningLevel.

        The value is a symbol in ['On', 'Off']. Default value is 'On'.
        """
        return self.get_attribute('PrintModelDetailsInMessages')

    def set_PrintModelDetailsInMessages(self, val):
        self._set_value_enum('PrintModelDetailsInMessages', val, _ON_OFF)

    PrintModelDetailsInMessages = property(get_PrintModelDetailsInMessages, set_PrintModelDetailsInMessages)


    def get_RandomSeed(self):
        """
        The search uses some randomization in some strategies. This parameter sets the seed of the random
        generator used by these strategies.

        The value is a non-negative integer. Default value is 0.
        """
        return self.get_attribute('RandomSeed')

    def set_RandomSeed(self, val):
        self._set_value_integer('RandomSeed', val)

    RandomSeed = property(get_RandomSeed, set_RandomSeed)


    def get_RelativeOptimalityTolerance(self):
        """
        This parameter sets a relative tolerance on the objective value for optimization models. This means
        that when CP Optimizer reports an optimal solution found, then there is no solution which improves
        the objective by more than the absolute value of the objective times the value of this parameter.
        The default value of this parameter is 1e-4. This parameter is used in conjunction with
        OptimalityTolerance. The optimality of a solution is proven if either of the two parameters'
        criteria are fulfilled.

        The value is a non-negative float. Default value is 0.0001.
        """
        return self.get_attribute('RelativeOptimalityTolerance', 0.0001)

    def set_RelativeOptimalityTolerance(self, val):
        self._set_value_float('RelativeOptimalityTolerance', val)

    RelativeOptimalityTolerance = property(get_RelativeOptimalityTolerance, set_RelativeOptimalityTolerance)


    def get_RestartFailLimit(self):
        """
        When SearchType is set to Restart, a depth-first search is restarted after a certain number of
        failures. This parameter controls the number of failures that must occur before restarting search.
        Possible values range from 0 to Infinity. The default value is 100. This value can increase after
        each restart: see the parameter RestartGrowthFactor.

        The value is an integer greater than 0. Default value is 100.
        """
        return self.get_attribute('RestartFailLimit', 100)

    def set_RestartFailLimit(self, val):
        self._set_value_integer('RestartFailLimit', val, 1)

    RestartFailLimit = property(get_RestartFailLimit, set_RestartFailLimit)


    def get_RestartGrowthFactor(self):
        """
        When SearchType is set to Restart, a depth-first search is restarted after a certain number of
        failures. This parameter controls the increase of this number between restarts. If the last fail
        limit was f after a restart, for next run, the new fail limit will be f times the value of this
        parameter. Possible values of this parameter range from 1.0 to Infinity. The default value is 1.05.
        The initial fail limit can be controlled with the parameter RestartFailLimit.

        The value is a float greater or equal to 1. Default value is 1.15.
        """
        return self.get_attribute('RestartGrowthFactor')

    def set_RestartGrowthFactor(self, val):
        self._set_value_float('RestartGrowthFactor', val, min=1)

    RestartGrowthFactor = property(get_RestartGrowthFactor, set_RestartGrowthFactor)


    def get_SearchType(self):
        """
        This parameter determines the type of search that is applied when solving a problem.

         * When set to *DepthFirst*, a regular depth-first search is applied.
         * When set to *Restart*, a depth-first search that restarts from time to time is applied.
         * When set to *IterativeDiving* on scheduling problems (ones with at least one interval variable),
           a more aggressive diving technique is applied in order to find solutions to large problems more quickly.
         * When set to *MultiPoint*, a method that combines a set of - possibly partial - solutions is applied.
         * When set to *Auto* in sequential mode, this value chooses the appropriate search method to be used.
           In general Auto will be the Restart search. The default value is Auto.

        In parallel mode (i.e, when the number of workers is greater than one - see the Workers
        parameter), the different searches described above are spread over the workers. When the value of
        SearchType is Auto, then the decision of choosing the search type for a worker is automatically
        made; otherwise, all workers execute the same type of search. Note that in the latter case, the
        workers will not do the same exploration due to some randomness introduced to break ties in decision
        making.

        The value is a symbol in ['DepthFirst', 'Restart', 'MultiPoint', 'IterativeDiving', 'Auto']. Default value is 'Auto'.
        """
        return self.get_attribute('SearchType')

    def set_SearchType(self, val):
        self._set_value_enum('SearchType', val, _SEARCH_TYPES)

    SearchType = property(get_SearchType, set_SearchType)


    def get_SequenceInferenceLevel(self):
        """
        This parameter specifies the inference level for every constraint Sequence extracted to the invoked
        CP instance. Possible values for this parameter are Default, Low, Basic, Medium, and Extended. The
        default value is Default, which allows the inference strength of all Sequence constraints to be
        controlled via DefaultInferenceLevel.

        The value is a symbol in ['Default', 'Low', 'Basic', 'Medium', 'Extended']. Default value is 'Default'.
        """
        return self.get_attribute('SequenceInferenceLevel')

    def set_SequenceInferenceLevel(self, val):
        self._set_value_enum('SequenceInferenceLevel', val, _INFERENCE_LEVELS)

    SequenceInferenceLevel = property(get_SequenceInferenceLevel, set_SequenceInferenceLevel)


    def get_SolutionLimit(self):
        """
        This parameter limits the number of feasible solutions that are found before terminating a search.

        The value is a non-negative integer, or None (default) that does not set any limit.
        """
        return self.get_attribute('SolutionLimit')

    def set_SolutionLimit(self, val):
        self._set_value_integer('SolutionLimit', val)

    SolutionLimit = property(get_SolutionLimit, set_SolutionLimit)


    def get_StateFunctionInferenceLevel(self):
        """
        This parameter specifies the inference level for constraints on state functions StateFunction
        extracted to the invoked CP instance. Possible values for this parameter are Default, Low, Basic,
        Medium, and Extended. The default value is Default, which allows the inference strength of all
        constraints on state functions StateFunction to be controlled via DefaultInferenceLevel.

        The value is a symbol in ['Default', 'Low', 'Basic', 'Medium', 'Extended']. Default value is 'Default'.
        """
        return self.get_attribute('StateFunctionInferenceLevel')

    def set_StateFunctionInferenceLevel(self, val):
        self._set_value_enum('StateFunctionInferenceLevel', val, _INFERENCE_LEVELS)

    StateFunctionInferenceLevel = property(get_StateFunctionInferenceLevel, set_StateFunctionInferenceLevel)


    def get_TemporalRelaxation(self):
        """
        This advanced parameter can be used to control the usage of a temporal relaxation internal to the
        invoking CP engine. This parameter can take values On or Off, with On being the default, meaning the
        relaxation is used in the engine when needed. For some models, using the relaxation becomes
        inefficient, and you may deactivate the use of the temporal relaxation using value Off.

        The value is a symbol in ['On', 'Off']. Default value is 'On'.
        """
        return self.get_attribute('TemporalRelaxation')

    def set_TemporalRelaxation(self, val):
        self._set_value_enum('TemporalRelaxation', val, _ON_OFF)

    TemporalRelaxation = property(get_TemporalRelaxation, set_TemporalRelaxation)


    def get_TimeLimit(self):
        """
        This parameter limits the CPU time spent solving before terminating a search. The time is given in
        seconds.

        The value is a non-negative float, or None (default) that does not set any limit.
        """
        return self.get_attribute('TimeLimit')

    def set_TimeLimit(self, val):
        self._set_value_float('TimeLimit', val)

    TimeLimit = property(get_TimeLimit, set_TimeLimit)


    def get_TimeMode(self):
        """
        This parameter defines how time is measured in CP Optimizer, the two legal values being ElapsedTime
        and CPUTime. CP Optimizer uses time for both display purposes and for limiting the search via
        TimeLimit. Note that when multiple processors are available and the number of workers (Workers) is
        greater than one, then the CPU time can be greater than the elapsed time by a factor up to the
        number of workers.

        The value is a symbol in ['CPUTime', 'ElapsedTime']. Default value is 'ElapsedTime'.
        """
        return self.get_attribute('TimeMode')

    def set_TimeMode(self, val):
        self._set_value_enum('TimeMode', val, (VALUE_CPU_TIME, VALUE_ELAPSED_TIME))

    TimeMode = property(get_TimeMode, set_TimeMode)


    def get_UseFileLocations(self):
        """
        This parameter controls whether CP Optimizer processes file locations. With each constraint,
        variable or expression it is possible to associate a source file location (file name and line
        number). CP Optimizer can use locations later for reporting errors and conflicts. Locations are also
        included in exported/dumped models (#line directives). Legal values for this parameter are On (the
        default) and Off. When the value is Off then CP Optimizer ignores locations in the input model and
        also does not export them in CPO file format (functions dumpModel and exportModel).

        The value is a symbol in ['On', 'Off']. Default value is 'On'.
        """
        return self.get_attribute('UseFileLocations')

    def set_UseFileLocations(self, val):
        self._set_value_enum('UseFileLocations', val, _ON_OFF)

    UseFileLocations = property(get_UseFileLocations, set_UseFileLocations)


    def get_WarningLevel(self):
        """
        This parameter controls the level of warnings issued by CP Optimizer when a solve is launched.
        Specifically, all warnings of level higher than this parameter are masked. Since CP Optimizer
        warning levels run from 1 to 3, setting this parameter to 0 turns off all warnings. Warnings issued
        may indicate potential errors or inefficiencies in your model. The default value of this parameter
        is 2. See also PrintModelDetailsInMessages.

        The value is an integer in [0..3]. Default value is 2.
        """
        return self.get_attribute('WarningLevel')

    def set_WarningLevel(self, val):
        self._set_value_integer('WarningLevel', val, min=0, max=3)

    WarningLevel = property(get_WarningLevel, set_WarningLevel)


    def get_Workers(self):
        """
        This parameter sets the number of workers to run in parallel to solve your model. If the number of
        workers is set to n (with n greater than one), the CP optimizer will create n workers, each in their
        own thread, that will work together to solve the problem. The emphasis of these workers is more to
        find better feasible solutions and then to speed up the proof of optimality. The default value is
        Auto. This amounts to using as many workers as there are CPU cores available on the machine. Note
        that the memory required by CP Optimizer grows roughly linearly as the number of workers is
        increased. If you are solving a very large model on a multi-core processor and memory usage is an
        issue, it is advisable to specify a reduced number of workers, or even one worker, rather than use
        the default value.

        The value is a positive integer. Default value is Auto.
        """
        return self.get_attribute('Workers')

    def set_Workers(self, val):
        assert (val is None) or (is_int(val) and val > 0) or val == "Auto", \
            "Value of parameter 'Workers' should be a positive integer or 'Auto'"
        self.set_attribute('Workers', val)

    Workers = property(get_Workers, set_Workers)


    def _set_value_enum(self, name, val, accepted):
        """ Check if a value is in a given set and set it if ok.
        Args:
            name:     Parameter name
            val:      Parameter value to set
            accepted: Accepted values
        """
        assert (val is None) or (val in accepted), "Value of parameter '{}' should be in {}".format(name, accepted)
        self.set_attribute(name, val)

    def _set_value_integer(self, name, val, min=0, max=None):
        """ Check if a value is a non-negative integer and set it if ok.
        Args:
            name:     Parameter name
            val:      Parameter value to set
            min:      (Optional) Minimun value
            max:      (Optional) Maximum value
        """
        assert (val is None) or \
               (is_int(val) and ((min is None or val >= min) and (max is None or val <= max))), \
                     "Value of parameter '{}' should be an integer in [{}..{}]"\
                         .format(name, "-Infinity" if min is None else min, "Infinity" if max is None else max)
        self.set_attribute(name, val)


    def _set_value_float(self, name, val, min=0, max=None):
        """ Check if a value is a non-negative float and set it if ok.
        Args:
            name:     Parameter name
            val:      Parameter value to set
            min:      (Optional) Minimun value
            max:      (Optional) Maximum value
        """
        assert (val is None) or \
               (is_number(val) and ((min is None or val >= min) and (max is None or val <= max))), \
                      "Value of parameter '{}' should be a float in [{}..{}]"\
                          .format(name, "-Infinity" if min is None else min, "Infinity" if max is None else max)
        self.set_attribute(name, val)




# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

import json
from docplex.mp.constants import ConflictStatus


class JSONConflictHandler(object):
    '''A class that provides utilities to process the JSON conflict from DOcplexcloud
    '''

    # json keys
    JSON_LINEAR_CTS_KEY = 'linearConstraints'

    row_conflict_by_status = {'excluded': ConflictStatus.Excluded,
                              'possible': ConflictStatus.Possible_member,
                              'member': ConflictStatus.Member}
    grp_conflict_by_status = row_conflict_by_status
    col_conflict_by_status = {'excluded': ConflictStatus.Excluded,
                              'possible': ConflictStatus.Possible_member,
                              'possible_lb': ConflictStatus.Possible_member_lower_bound,
                              'possible_ub': ConflictStatus.Possible_member_upper_bound,
                              'member': ConflictStatus.Member,
                              'member_lb': ConflictStatus.Member_lower_bound,
                              'member_ub': ConflictStatus.Member_upper_bound}

    def __init__(self, conflict_string, grps_dict, has_conflict=None):
        """Initialize a new JSONConflictHandler

        This handler is initialized with a json fragment with the CPLEXConflict.

        Args:
            conflict_string: The json text containing a CPLEXSolution
            grps_dict: dictionary of groups provided to the DOcplexcloud job, with keys = groups index
            has_conflict: if set to True or False, forces the has_conflict status. If None, this
                is True if json is not None.
        """
        self.has_conflict = bool(conflict_string) if has_conflict is None else has_conflict
        if conflict_string is not None:
            self.json = json.loads(conflict_string, parse_constant='utf-8')['CPLEXConflict']
        else:
            self.json = None

        #
        self._grps_dict = grps_dict
        # used to store data from the json
        self.__cols = None
        self.__rows = None
        self.__grps = None

    def _get_grps(self):
        if self.__grps is None:
            self.__grps = self.json.get('grp', [])
        return self.__grps

    def get_conflict_grps_list(self):
        all_grps = self._get_grps()
        conflict_grps_list = [(int(g['index']),
                               self._grps_dict[int(g['index'])],
                               self.grp_conflict_by_status.get(g['status'], None))
                              for g in all_grps]
        return conflict_grps_list

    def _get_cols(self):
        if self.__cols is None:
            self.__cols = self.json.get('col', [])
        return self.__cols

    def _get_rows(self):
        if self.__rows is None:
            self.__rows = self.json.get('row', [])
        return self.__rows



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------
# Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis

"""
This module allows to solve a model expressed as a CPO file using DOcplexcloud services.
"""

import docplex.cp.solution as solution
from docplex.cp.solution import CpoProcessInfos, CpoSolverInfos, CpoSolveResult, CpoRefineConflictResult
from docplex.cp.solver.docloud_client import JobClient
from docplex.cp.utils import CpoException, is_number
import docplex.cp.solver.solver as solver
import time
import warnings


###############################################################################
##  Constants
###############################################################################

# Solution status indication that a json result is expected
_STATUS_WITH_RESULT = (solution.SOLVE_STATUS_FEASIBLE,
                       solution.SOLVE_STATUS_OPTIMAL)


###############################################################################
##  Public classes
###############################################################################

class CpoSolverDocloud(solver.CpoSolverAgent):
    """ Solver of CPO model using DOcplexcloud services. """
    
    def __init__(self, solver, params, context):
        """ Create a new solver using DOcplexcloud web service

        Args:
            solver:   Parent solver
            params:   Solving parameters
            context:  DOcplexcloud Solver context
        Raises:
            CpoException if jar file does not exists
        """
        warnings.warn(
            'Solve using \'docloud\' agent is deprecated. Consider submitting your model to DOcplexcloud. See https://ibm.biz/BdYhhK',
            DeprecationWarning)
        if (context.key is None) or (' ' in context.key):
            raise CpoException("Your DOcplexcloud key has not been set")
        super(CpoSolverDocloud, self).__init__(solver, params, context)


    def solve(self):
        """ Solve the model

        According to the value of the context parameter 'verbose', the following information is logged
        if the log output is set:
         * 1: Total time spent to solve the model
         * 2: Calls to DOcplexcloud solving
         * 3: Detailed DOcplexcloud job information
         * 4: REST requests and response codes

        Returns:
            Model solve result,
            object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        Raises:
            CpoException if error occurs
        """
        return self._submit_job('Solve', CpoSolveResult)


    def refine_conflict(self):
        """ This method identifies a minimal conflict for the infeasibility of the current model.

        See documentation of CpoSolver.refine_conflict() for details.

        Returns:
            Conflict result,
            object of class :class:`~docplex.cp.solution.CpoRefineConflictResult`.
        """
        # Ensure cpo model is generated with all constraints named
        self.context.model.name_all_constraints = True

        return self._submit_job('RefineConflict', CpoRefineConflictResult)


    def propagate(self):
        """ This method invokes the propagation on the current model.

        See documentation of CpoSolver.propagate() for details.

        Returns:
            Propagation result,
            object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        """
        return self._submit_job('Propagate', CpoSolveResult)


    def _submit_job(self, cmd, rclass):
        """ Submit a solving job with given command.

        According to the value of the context parameter 'verbose', the following information is logged
        if the log output is set:
         * 1: Total time spent to solve the model
         * 2: Calls to DOcplexcloud solving
         * 3: Detailed DOcplexcloud job information
         * 4: REST requests and response codes

        Args:
            cmd:    Solving command, in (Solve, RefineConflict, Propagate)
            rclass: Result object class
        Returns:
            Model solve result (object of class CpoSolveResult)
        Raises:
            CpoException if error occurs
        """
        # Create DOcplexcloud client
        ctx = self.context
        client = JobClient(ctx)

        # Convert model into CPO format
        cpostr = self._get_cpo_model_string()

        # Encode model
        stime = time.time()
        cpostr = cpostr.encode('utf-8')
        self.process_infos[CpoProcessInfos.TOTAL_UTF8_ENCODE_TIME] = time.time() - stime

        # Solve model and retrieve solution
        name = self.model.get_name()
        if is_number(ctx.params.TimeLimit):
            maxwait = ctx.params.TimeLimit + ctx.request_timeout + ctx.result_wait_extra_time
        else:
            maxwait = 0
        try:
            # Create job and start execution
            stime = time.time()
            client.create_cpo_job(name, cpostr, cmd)
            self.process_infos[CpoProcessInfos.TOTAL_DATA_SEND_TIME] = time.time() - stime

            # Start execution
            client.execute_job()

            # Wait job termination
            lgntf = None
            if self.log_enabled:
                lgntf = (lambda recs: self._add_log_data(''.join(r['message'] + "\n" for r in recs)))
            client.wait_job_termination(maxwait=maxwait, lognotif=lgntf)
            jinfo = client.get_info()

            # Trace response if required
            if ctx.is_log_enabled(3):
                ctx.log(3, "Job info:")
                for k in jinfo.keys():
                    ctx.log(3, k, " : ", jinfo[k])

            # Check failure
            fail = jinfo.get('failure', None)
            if fail is not None:
                raise CpoException(fail.get('message', "Unknown failure"))

            # Get solve status
            ssts = _get_solve_status(jinfo)
            if (cmd == "Solve") and (ssts not in _STATUS_WITH_RESULT):
                jsol = None
            else:
                # Get solution
                try:
                    stime = time.time()
                    jsol = client.get_attachment("solution.json")
                    self.process_infos[CpoProcessInfos.RESULT_RECEIVE_TIME] = time.time() - stime
                    self.process_infos[CpoProcessInfos.RESULT_DATA_SIZE] = len(jsol)
                    stime = time.time()
                    jsol = jsol.decode('utf-8')
                    self.process_infos[CpoProcessInfos.TOTAL_UTF8_DECODE_TIME] = time.time() - stime
                except Exception as e:
                    raise CpoException("Model solution access error: " + str(e))

            # Create solution structure
            msol = self._create_result_object(rclass, jsol)
            if not jsol:
                msol.solve_status = ssts
            sinfos = msol.solver_infos

            # Add solve details if no json solution
            detls = jinfo.get('details', None)
            if detls is not None:
                # Retrieve common infos
                for k in detls:
                    if not(('.' in k) or ('_' in k)):
                        sinfos[k] = _value_of(detls[k])
                # Retrieve special hard-coded values
                if isinstance(msol, CpoSolveResult) and msol.solution.objective_values is None:
                    oval = detls.get('PROGRESS_CURRENT_OBJECTIVE', None)
                    if oval is not None:
                        msol.solution.objective_values = [float(v) for v in oval.split(";")]
                sinfos.setdefault(CpoSolverInfos.NUMBER_OF_CONSTRAINTS,        int(detls.get('MODEL_DETAIL_CONSTRAINTS', 0)))
                sinfos.setdefault(CpoSolverInfos.NUMBER_OF_INTEGER_VARIABLES,  int(detls.get('MODEL_DETAIL_INTEGER_VARS', 0)))
                sinfos.setdefault(CpoSolverInfos.NUMBER_OF_INTERVAL_VARIABLES, int(detls.get('MODEL_DETAIL_INTERVAL_VARS', 0)))
                sinfos.setdefault(CpoSolverInfos.NUMBER_OF_SEQUENCE_VARIABLES, int(detls.get('MODEL_DETAIL_SEQUENCE_VARS', 0)))

            # Add solve time
            if CpoSolverInfos.SOLVE_TIME not in sinfos:
                sinfos[CpoSolverInfos.SOLVE_TIME] = (float(jinfo.get('endedAt', 0)) - float(jinfo.get('startedAt', 0))) / 1000

        finally:
            # Delete job
            if ctx.clean_job_after_solve:
                client.clean_job()

        # Return
        return msol



###############################################################################
##  Model solution building functions
###############################################################################


def _get_solve_status(jinfo):
    """ Build solve status from docloud job info
    Args:
        jinfo:  DOcloud job info
    """
    sts = jinfo.get('executionStatus', None)
    if sts == "INTERRUPTED":         return solution.SOLVE_STATUS_JOB_ABORTED
    if sts == "FAILED":              return solution.SOLVE_STATUS_JOB_FAILED

    sts = jinfo.get('solveStatus', None)
    if sts == "FEASIBLE_SOLUTION":   return solution.SOLVE_STATUS_FEASIBLE
    if sts == "INFEASIBLE_SOLUTION": return solution.SOLVE_STATUS_INFEASIBLE
    if sts == "OPTIMAL_SOLUTION":    return solution.SOLVE_STATUS_OPTIMAL

    return solution.SOLVE_STATUS_UNKNOWN


def _value_of(vstr):
    """ Build a python value from a string
    Args:
        vstr:  Value as a string
    Returns:
        Integer, float or string value
    """
    try:
        return int(vstr)
    except Exception:
        try:
            return float(vstr)
        except Exception:
            return vstr




# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

from __future__ import print_function
from docplex.mp.utils import DOcplexException, DOcplexLimitsExceeded, resolve_pattern, is_int, is_string

from enum import Enum

##########################
# Error handling


class IErrorHandler(object):
    def __init__(self):
        pass  # pragma: no cover

    def info(self, msg, args=None):
        pass  # pragma: no cover

    def warning(self, msg, args=None):
        pass  # pragma: no cover

    def error(self, msg, args=None):
        pass  # pragma: no cover

    def fatal(self, msg, args=None):
        pass  # pragma: no cover

    def ok(self):
        return False  # pragma: no cover

    def get_output_level(self):
        return 0  # pragma: no cover

    def set_output_level(self, new_level):
        pass  # pragma: no cover

    def ensure(self, condition, msg, *args):
        if not condition:
            self.fatal(msg, args)


class InfoLevel(Enum):
    """ Enumerated type for the possible output levels.

    Info levels are sorted in increasing order of severity: `INFO`, `WARNING`, `ERROR`, `FATAL`.
    Setting a level enables the printing of all messages from that severity and above.

    Example:
        Setting the level to `WARNING` enables the printing of `WARNING`, `ERROR`, and `FATAL` messages, but not
        `INFO` level messages.
        Setting the level to `FATAL` suppresses all messages, except for fatal errors.
    """
    INFO, WARNING, ERROR, FATAL = 1, 10, 100, 9999999

    @classmethod
    def parse(cls, arg, default_level=INFO):
        # INTERNAL
        if not arg:
            return cls.INFO
        elif isinstance(arg, cls):
            return arg
        elif is_string(arg):
            return cls._name2level_map().get(arg.lower(), default_level)
        elif is_int(arg):
            if arg < 10:
                # anything below 10 is INFO
                return cls.INFO
            elif arg < 100:
                return cls.WARNING
            elif arg < 1000:
                return cls.ERROR
            else:
                # level fatal prints nothing except fatal errors
                return cls.FATAL
        else:
            raise DOcplexException("Cannot convert this to InfoLevel: {0!r}".format(arg))

    def __str__(self):
        return self.name

    @staticmethod
    def _headers():
        return {InfoLevel.FATAL: "FATAL",
                InfoLevel.INFO: "*",
                InfoLevel.WARNING: "Warning:",
                InfoLevel.ERROR: "Error:"
                }

    @staticmethod
    def _name2level_map():
        return {"fatal": InfoLevel.FATAL,
                "error": InfoLevel.ERROR,
                "warning": InfoLevel.WARNING,
                "info": InfoLevel.INFO}

    def header(self):
        # cannot put the dict in the class
        # as it willbe interpreted as another enum value.
        return self._headers().get(self, "???")


class AbstractErrorHandler(IErrorHandler):
    TRACE_HEADER = "--"

    def __init__(self, output_level=InfoLevel.INFO):
        IErrorHandler.__init__(self)
        self._trace_enabled = False

        self._number_of_errors = 0
        self._number_of_warnings = 0
        self._number_of_fatals = 0
        self._output_level = InfoLevel.INFO
        self._is_print_suspended = False
        self._postponed = []
        self.set_output_level(output_level)

    @property
    def number_of_warnings(self):
        """ Returns the number of warnings.
        """
        return self._number_of_warnings

    @property
    def number_of_errors(self):
        """ Returns the number of errors.
        """
        return self._number_of_errors

    @property
    def number_of_fatals(self):
        return self._number_of_fatals

    def get_output_level(self):
        return self._output_level

    def set_output_level(self, output_level_arg):
        output_level = InfoLevel.parse(output_level_arg)
        if output_level != self._output_level:
            self._output_level = output_level

    def set_trace_mode(self, trace_mode):
        self._trace_enabled = trace_mode

    def enable_trace(self):
        self.set_trace_mode(True)

    def disable_trace(self):
        self.set_trace_mode(False)

    def is_trace_enabled(self):
        return self._trace_enabled

    def set_quiet(self):
        """ Changes the output level to enable only error messages.
        """
        self.set_output_level(InfoLevel.ERROR)

    def reset(self):
        self._number_of_errors = 0
        self._number_of_warnings = 0
        self._number_of_fatals = 0

    def _internal_is_printed(self, level):
        return self._output_level.value <= level.value

    def _internal_print_if(self, level, msg, args):
        if self._internal_is_printed(level):
            self._internal_print(level, msg, args)

    def _internal_print(self, level, msg, args):
        # resolve message w/ args
        header = level.header()
        self._internal_print_header(header, msg, args)

    def _internal_print_header(self, header, msg, args):
        resolved_message = resolve_pattern(msg, args)
        mline = '%s %s' % (header, resolved_message)
        if self._is_print_suspended:
            self._postponed.append(mline)
        else:
            print(mline)

    def trace_header(self):
        return self.TRACE_HEADER

    def trace(self, msg, args=None):
        if self.is_trace_enabled():
            self._internal_print_header(self.trace_header(), msg, args)

    def info(self, msg, args=None):
        self._internal_print_if(InfoLevel.INFO, msg, args)

    def warning(self, msg, args=None):
        self._number_of_warnings += 1
        self._internal_print_if(InfoLevel.WARNING, msg, args)

    def error(self, msg, args=None):
        docplex_error_stop_here()
        self._number_of_errors += 1
        self._internal_print_if(InfoLevel.ERROR, msg, args)

    def fatal(self, msg, args=None):
        self._number_of_fatals += 1
        resolved_message = resolve_pattern(msg, args)
        docplex_error_stop_here()
        raise DOcplexException(resolved_message)

    def fatal_limits_exceeded(self):
        docplex_error_stop_here()
        raise DOcplexLimitsExceeded()


    def ok(self):
        """ Checks whether the handler has not recorded any error.
        """
        return self._number_of_errors == 0 and self._number_of_fatals == 0

    def prints_trace(self):
        return self.is_trace_enabled()

    def prints_info(self):
        return self._internal_is_printed(InfoLevel.INFO)

    def prints_warning(self):
        return self._internal_is_printed(InfoLevel.WARNING)

    def prints_error(self):
        return self._internal_is_printed(InfoLevel.ERROR)

    def suspend(self):
        self._is_print_suspended = True

    def flush(self):
        self._is_print_suspended = False
        for m in self._postponed:
            print(m)
        self._postponed = []


def docplex_error_stop_here():
    # INTERNAL, use to set breakpoints
    pass


def docplex_add_trivial_infeasible_ct(ct):
    # INTERNAL: set breakpoint here to inspect ct
    pass

def docplex_fatal(msg, *args):
    resolved_message = resolve_pattern(msg, args)
    docplex_error_stop_here()
    raise DOcplexException(resolved_message)


class DefaultErrorHandler(AbstractErrorHandler):
    """ The default error handler class.

    """

    def __init__(self, output_level=InfoLevel):
        AbstractErrorHandler.__init__(self, output_level)


class SilentErrorHandler(AbstractErrorHandler):
    def __init__(self, output_level=InfoLevel):
        AbstractErrorHandler.__init__(self, output_level)

    def _internal_print(self, level, msg, args):
        # nothing out, this is the point!
        pass

    def suspend(self):
        pass

    def flush(self):
        pass





# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore
from six import iteritems


class _SolveHook(object):
    # INTERNAL
    def __init__(self):
        pass  # pragma: no cover

    def notify_start_solve(self, mdl, model_statistics):
        """ Notifies the start of a solve.

        Args:
            model_statistics: A dictionary of string->values with various data attributes of the model.

        """
        pass  # pragma: no cover

    def notify_end_solve(self, mdl, has_solution, status, obj, var_value_dict):
        """ Notifies the end of a solve.

        Args:
            has_solution: Boolean, True if solve returned a solution.
            status: An enumerated value of type JobSolveStatus.
            obj: The objective value if solved ok, else irrelevant.
            var_value_dict: A dictionary of variable names to values in the solution.
        """
        pass  # pragma: no cover


class TraceSolveHook(_SolveHook):
    def notify_start_solve(self, mdl, stats):
        print("-> start solve")
        stats.print_information()

    def notify_end_solve(self, mdl, has_solution, status, obj, var_value_dict):
        if has_solution:
            print("<- solve succeeds, status={0}, obj={1}".format(status, obj))
            for vn, vv in iteritems(var_value_dict):
                print("  - var \"{0:s}\" = {1!s}".format(vn, vv))
        else:
            print("<- solve fails, status={0}".format(status))



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

from enum import Enum

from docplex.mp.error_handler import docplex_fatal
from docplex.mp.utils import is_string, is_int

import operator


class VarBoundType(Enum):
    """This enumerated class renages over the two types of variable bounds:
        - LB is for lower bound
        - UB is for uupper bound

    This enumerated type is used in conflict refiner.
    """
    LB = 0
    UB = 1


class ComparisonType(Enum):
    """This enumerated class defines the various types of linear constraints:

        - LE for e1 <= e2 constraints

        - EQ for e1 == e2 constraints

        - GE for e1 >= e2 constraints

        where e1 and e2 denote linear expressions.
    """
    LE = 1, '<=', 'L', operator.le
    EQ = 2, '==', 'E', operator.eq
    GE = 3, '>=', 'G', operator.ge

    def __new__(cls, code, operator_symbol, cplex_kw, python_op):
        obj = object.__new__(cls)
        # predefined
        obj._value_ = code
        obj._cplex_code = cplex_kw
        obj._op_symbol = operator_symbol
        obj._pyop = python_op
        return obj

    # NOTE: Never add a static field in an enum class: it would be interpreted as an other enum

    @property
    def short_name(self):
        return self.name

    @property
    def cplex_code(self):
        return self._cplex_code

    @property
    def operator_symbol(self):
        """ Returns a string operator for the constraint.

        Example:
            Returns string "<=" for a e1 <= e2 constraint.

        Returns:
            string: A string describing the logical operator used in the constraint.
        """
        return self._op_symbol

    @property
    def python_operator(self):
        return self._pyop

    @classmethod
    def parse(cls, arg, do_raise=True):
        # INTERNAL
        # noinspection PyTypeChecker
        for op in cls:
            if arg == op or arg == op.value:
                return op
            elif is_string(arg):
                if arg == op._cplex_code \
                        or arg == str(op.value) \
                        or arg.lower() == op.name.lower():
                    return op
        else:
            if do_raise:
                docplex_fatal('cannot convert this to a comparison type: {0!r}'.format(arg))
            else:
                return None

    @classmethod
    def cplex_ctsense_to_python_op(cls, cpx_sense):
        return cls.parse(cpx_sense).python_operator

    @classmethod
    def almost_compare(cls, lval, op, rval, eps):
        if op is cls.LE:
            # lval <= rval with eps tolerance means lval-rval <= e
            return lval - rval <= eps
        elif op is cls.GE:
            # lval >= rval with eps tolerance means lval-rval >= -eps
            return lval - rval >= -eps
        elif op is cls.EQ:
            return abs(lval - rval) <= eps
        else:
            raise TypeError

    @classmethod
    def almost_equal(cls, lval, rval, eps):
        return cls.almost_compare(lval, cls.EQ, rval, eps)


class RelaxationMode(Enum):
    """ This enumerated type describes the different strategies for model relaxation: MinSum, OptSum, MinInf, OptInf, MinQuad, OptQuad.

    Relaxation algorithms work in two phases: In the first phase, they attempt to find a
    feasible solution while making minimal changes to the model (according to a metric). In the second phase, they
    attempt to find an optimal solution while keeping the relaxation at the minimal value found in phase 1.

    Values of this type define two aspects of the algorithm:
        - whether or not they should continue to a second phase: all OptXxx values continue to a
          second phase, and MinXxx values stop at phase 1.
        - which metric to use for evaluating the relaxation in the first phase. There are three metrics:
              - the sum of relaxations for OptSum, MinSum, or
              - the total number of constraints being relaxed for OptInf, MinInf, or
              - the sum of squares of relaxations for OptQuad, MinQuad.

    """

    MinSum, OptSum, MinInf, OptInf, MinQuad, OptQuad = range(6)

    @staticmethod
    def parse(arg):
        # INTERNAL
        # noinspection PyTypeChecker
        for m in RelaxationMode:
            if arg == m or arg == m.value:
                return m
            elif is_string(arg):
                if arg == str(m.value) or arg.lower() == m.name.lower():
                    return m
        else:
            docplex_fatal('cannot parse this as a relaxation mode: {0!r}'.format(arg))

    @staticmethod
    def get_no_optimization_mode(mode):
        assert isinstance(mode, RelaxationMode)
        # even values are MinXXX modes
        relax_code = mode.value
        if 0 == relax_code % 2:
            return mode
        else:
            # OptXXX is 2k+1 when MinXXX is 2k
            return RelaxationMode(relax_code - 1)

    def __repr__(self):
        return 'docplex.mp.RelaxationMode.{0}'.format(self.name)


class ConflictStatus(Enum):
    """
    This enumerated class defines the conflict status types.
    """
    Excluded, Possible_member, Possible_member_lower_bound, Possible_member_upper_bound, \
      Member, Member_lower_bound, Member_upper_bound = -1, 0, 1, 2, 3, 4, 5


class SOSType(Enum):
    """This enumerated class defines the SOS types:

        - SOS1 for SOS type 1

        - SOS1 for SOS type 2.
    """
    SOS1, SOS2 = 1, 2

    def lower(self):
        return self.name.lower()

    @staticmethod
    def parse(arg,
              sos1_tokens=frozenset(['1', 'sos1']),
              sos2_tokens=frozenset(['2', 'sos2'])):
        if isinstance(arg, SOSType):
            return arg
        elif 1 == arg:
            return SOSType.SOS1
        elif 2 == arg:
            return SOSType.SOS2
        elif is_string(arg):
            arg_lower = arg.lower()
            if arg_lower in sos1_tokens:
                return SOSType.SOS1
            elif arg_lower in sos2_tokens:
                return SOSType.SOS2

        docplex_fatal("Cannot convert to SOS type: {0!s} - expecting 1|2|'sos1'|'sos2'", arg)

    def _cpx_sos_type(self):
        # INTERNAL
        return str(self.value)

    @property
    def size(self):
        return self.value

    def __repr__(self):
        return 'docplex.mp.SOSType.{0}'.format(self.name)


class SolveAttribute(Enum):
    duals = 1, False, True
    slacks = 2, False, True
    reduced_costs = 3, True, True

    def __new__(cls, code, is_for_vars, requires_solve):
        obj = object.__new__(cls)
        # predefined
        obj._value_ = code
        obj.is_var_attribute = is_for_vars
        obj.requires_solve = requires_solve
        return obj

    @classmethod
    def parse(cls, arg, do_raise=True):
        # INTERNAL
        # noinspection PyTypeChecker
        for m in cls:
            if arg == m or arg == m.value:
                return m
            elif is_string(arg):
                if arg == str(m.value) or arg.lower() == m.name.lower():
                    return m
        else:
            if do_raise:
                docplex_fatal('cannot convert this to a solve attribute: {0!r}'.format(arg))
            else:
                return None


class UpdateEvent(Enum):
    # INTERNAL
    NoOp = 0
    #
    # Linear constraint events
    LinearConstraintCoef = 1
    LinearConstraintRhs = 2
    LinearConstraintGlobal = 3  # logical and of Coef + Rhs
    ConstraintSense = 4

    # Range constraint events
    RangeConstraintBounds = 5
    RangeConstraintExpr = 6

    # Expression events
    ExprConstant = 8
    LinExprCoef = 16
    LinExprGlobal = 24
    LinExprPromotedToQuad = 25  # objective is ok, constraint wont support this.

    # Quad
    QuadExprQuadCoef = 32
    QuadExprGlobal = 64

    # Ind
    IndicatorLinearConstraint = 128

    # Quadct
    QuadraticConstraintGlobal = 256

    def __bool__(self):
        return True if self.value else False


class ObjectiveSense(Enum):
    """
    This enumerated class defines the two types of objectives, `Minimize` and `Maximize`.
    """
    Minimize, Maximize = 1, 2

    def is_minimize(self):
        return self is ObjectiveSense.Minimize

    def is_maximize(self):
        return self is ObjectiveSense.Maximize

    @property
    def cplex_coef(self):
        return 1 if self.is_minimize() else -1

    @property
    def verb(self):
        # INTERNAL
        return self.name.lower()

    @property
    def short_name(self):
        return self.verb[:3]

    @staticmethod
    def parse(arg, logger=None, default_sense=None):
        if isinstance(arg, ObjectiveSense):
            return arg

        elif is_string(arg):
            lower_text = arg.lower()
            if lower_text in {"minimize", "min"}:
                return ObjectiveSense.Minimize
            elif lower_text in {"maximize", "max"}:
                return ObjectiveSense.Maximize
            elif default_sense:
                logger.error(
                    "Text is not recognized as objective sense: {0}, expecting \"min\" or \"max\" - using default {1:s}",
                    (arg, default_sense))
                return default_sense
            elif logger:
                logger.fatal("Text is not recognized as objective sense: {0}, expecting ""min"" or ""max", (arg,))
            else:
                docplex_fatal("Text is not recognized as objective sense: {0}, expecting ""min"" or ""max".format(arg))
        elif is_int(arg):
            if arg == 1:
                return ObjectiveSense.Minimize
            elif -1 == arg:
                return ObjectiveSense.Maximize
            else:
                logger.fatal("cannot convert: <{}> to objective sense", (arg,))
        elif arg is None:
            return default_sense
        elif logger:
            logger.fatal("cannot convert: <{}> to objective sense", (arg,))
        else:
            docplex_fatal("cannot convert: <{}> to objective sense".format(arg))


# noinspection PyPep8
class CplexScope(Enum):
    def __new__(cls, code, prefix, descr):
        obj = object.__new__(cls)
        # predefined
        obj._value_ = code
        obj.prefix = prefix
        obj.descr = descr
        return obj

    # INTERNAL
    VAR_SCOPE = 0, 'x', 'variables'
    LINEAR_CT_SCOPE = 1, 'c', 'linear constraints'
    IND_CT_SCOPE = 2, 'ic', 'indicators'
    QUAD_CT_SCOPE = 3, 'qc', 'quadratic constraints'
    PWL_CT_SCOPE = 4, 'pwl', 'piecewise constraints'
    SOS_SCOPE = 5, 'sos', 'SOS'


class QualityMetric(Enum):
    def __new__(cls, code, has_int, cpx_codename):
        obj = object.__new__(cls)
        # predefined
        obj._value_ = code
        obj.has_int = has_int
        obj.codename = cpx_codename
        return obj

    @property
    def cpx_codename(self):
        return 'CPX_' + self.codename

    @property
    def code(self):
        return self._value_

    @property
    def key(self):
        return self.codename.lower()

    @property
    def int_key(self):
        return '%s.int' % self.codename.lower()

    max_primal_infeasibility = 1, 1, 'MAX_PRIMAL_INFEAS'
    max_scaled_primal_infeasibility = 2, 1, 'MAX_SCALED_PRIMAL_INFEAS'
    sum_primal_infeasibilities = 3, 0, 'SUM_PRIMAL_INFEAS'
    sum_scaled_primal_infeasibilities = 4, 0, 'SUM_SCALED_PRIMAL_INFEAS'

    max_dual_infeasibility = 5, 1, 'MAX_DUAL_INFEAS'
    max_scaled_dual_infeasibility = 6, 1, 'MAX_SCALED_DUAL_INFEAS'
    sum_dual_infeasibilities = 7, 0, 'SUM_DUAL_INFEAS'
    sum_scaled_dual_infeasibilities = 8, 0, 'SUM_SCALED_DUAL_INFEAS'

    max_int_infeasibility = 9, 1, 'MAX_INT_INFEAS'
    sum_integer_infeasibilities = 10, 0, 'SUM_INT_INFEAS'

    max_primal_residual = 11, 1, 'MAX_PRIMAL_RESIDUAL'
    max_scaled_primal_residual = 12, 1, 'MAX_SCALED_PRIMAL_RESIDUAL'
    sum_primal_residual = 13, 0, 'SUM_PRIMAL_RESIDUAL'
    sum_scaled_primal_residual = 14, 0, 'SUM_SCALED_PRIMAL_RESIDUAL'

    max_dual_residual = 15, 1, 'MAX_DUAL_RESIDUAL'
    max_scaled_dual_residual = 16, 1, 'MAX_SCALED_DUAL_RESIDUAL'
    sum_dual_residual = 17, 0, 'SUM_DUAL_RESIDUAL'
    sum_scaled_dual_residual = 18, 0, 'SUM_SCALED_DUAL_RESIDUAL'

    max_comp_slack = 19, 1, 'MAX_COMP_SLACK'  # gap here
    sum_comp_slack = 21, 0, 'SUM_COMP_SLACK'  # gap here

    max_x = 23, 1, 'MAX_X'
    max_scaled_x = 24, 1, 'MAX_SCALED_X'
    max_pi = 25, 1, 'MAX_PI'
    max_scaled_pi = 26, 1, 'MAX_SCALED_PI'
    max_slack = 27, 1, 'MAX_SLACK'
    max_scaled_slack = 28, 1, 'MAX_SCALED_SLACK'

    max_reduced_cost = 29, 1, 'MAX_RED_COST'
    max_scaled_reduced_cost = 30, 1, 'MAX_SCALED_RED_COST'

    sum_x = 31, 0, 'SUM_X'
    sum_scaled_x = 32, 0, 'SUM_SCALED_X'
    sum_pi = 33, 0, 'SUM_PI'
    sum_scaled_pi = 34, 0, 'SUM_SCALED_PI'
    sum_slack = 35, 0, 'SUM_SLACK'
    sum_scaled_slack = 36, 0, 'SUM_SCALED_SLACK'

    sum_reduced_cost = 37, 0, 'SUM_RED_COST'
    sum_scaled_reduced_cost = 38, 0, 'SUM_SCALED_RED_COST'

    kappa = 39, 0, 'KAPPA'
    objective_gap = 40, 0, 'OBJ_GAP'
    dual_objective = 41, 0, 'DUAL_OBJ'
    primal_objective = 42, 0, 'PRIMAL_OBJ'

    max_quadratic_primal_residual = 43, 1, 'MAX_QCPRIMAL_RESIDUAL'
    sum_quadratic_primal_residual = 44, 0, 'SUM_QCPRIMAL_RESIDUAL'
    max_quadratic_slack_infeasibility = 45, 1, 'MAX_QCSLACK_INFEAS'
    sum_quadratic_slack_infeasibility = 46, 0, 'SUM_QCSLACK_INFEAS'
    max_quadratic_slack = 47, 1, 'MAX_QCSLACK'
    sum_quadratic_slack = 48, 0, 'SUM_QCSLACK'
    max_indicator_slack_infeasibility = 49, 1, 'MAX_INDSLACK_INFEAS'
    sum_indicator_slack_infeasibility = 50, 0, 'SUM_INDSLACK_INFEAS'

    exact_kappa = 51, 0, 'EXACT_KAPPA'
    kappa_stable = 52, 0, 'KAPPA_STABLE'
    kappa_suspicious = 53, 0, 'KAPPA_SUSPICIOUS'
    kappa_unstable = 54, 0, 'KAPPA_UNSTABLE'
    kappa_illposed = 55, 0, 'KAPPA_ILLPOSED'
    kappa_max = 56, 0, 'KAPPA_MAX'
    kappa_attention = 57, 0, 'KAPPA_ATTENTION'

    @classmethod
    def parse(cls, txt, raise_on_error=True):
        for qm in cls:
            if txt == qm.name:
                return qm
            elif txt == qm.value:
                return qm
            elif txt == qm.cpx_codename:
                return qm
        else:
            fmt = '* cannot interpret this as a QualityMetric enum: {0!r}'
            if raise_on_error:
                docplex_fatal(fmt, txt)
            else:
                print(fmt.format(txt))
                return None


class BasisStatus(Enum):
    """ This enumerated type describes the different values for basis status.

    Basis status can be queried for variables and linear constraints in LP problems.

    Possible values are:

    - NotABasisStatus: invalid or unknown status,
    - Basic, means the variable belongs to the base,
    - AtLower, means the variable is non-basic, at its lower bound,
    - AtUpper, means the variable is non-basic, at its upper bound,
    - FreeNonBasic, means the variable is nonbasic and is not at a bound.

    See Also:
        The different values for basis status can be found in the CPLEX documentation:

        https://www.ibm.com/support/knowledgecenter/SSSA5P_12.9.0/ilog.odms.cplex.help/refcallablelibrary/cpxapi/getbase.html

    """

    def __new__(cls, code, cpx_codename):
        obj = object.__new__(cls)
        # predefined
        obj._value_ = code
        obj.codename = cpx_codename
        return obj

    NotABasisStatus = -1, "NotBasisStatus"
    AtLowerBound = 0, "CPX_AT_LOWER"
    Basic = 1, "CPX_BASIC"
    AtUpperBound = 2, "CPX_AT_UPPER"
    FreeNonBasic = 3, "CPX_FREE_SUPER"

    @classmethod
    def parse(cls, code):
        for bs in cls:
            if bs.value == code:
                return bs
        else:
            return cls.NotABasisStatus


class WriteLevel(Enum):
    """
    This enumerated class controls what is written in MST mip start files.
    The numeric value is identical to the CPLEX WriteLevel parameter values.

    The possible values are (in order of decreasing quantity of information written).

        - AllVars: all variables are written
        - DiscreteVars: all discrete variables are  written (binary, integer, semi-integer)
        - NonZeroVars: all non-zero vars are written, regardless of their type.
        - DiscreteNonZeroVars: all discrete non-zero vars are written.

        - Auto: automatic value, same as DiscreteVars.

    *New in version 2.10*
    """

    def __new__(cls, code, short_name):
        obj = object.__new__(cls)
        # predefined
        obj._value_ = code
        obj.short_name = short_name
        return obj

    Auto = 0, "auto"  # same as DiscreteVars: filter discrete, keep zeros
    AllVars = 1, "all"  # write all variables and their value, zero or nonzero
    DiscreteVars = 2, "discrete"  # write all discrete variables and their value
    NonZeroVars = 3, "nonzero"  # write only nonzero variables
    NonZeroDiscreteVars = 4, "nonzero_discrete"  # write nonzero discrete variables

    def filter_zeros(self):
        return self in {WriteLevel.NonZeroVars, WriteLevel.NonZeroDiscreteVars}

    def filter_discrete(self):
        return self in {WriteLevel.Auto, WriteLevel.DiscreteVars, WriteLevel.NonZeroDiscreteVars}

    @classmethod
    def parse(cls, level):
        if level is None:
            return cls.Auto
        elif isinstance(level, cls):
            return level
        else:
            for wl in cls:
                if wl.value == level:
                    return wl
                elif is_string(level):
                    llevel = level.lower()
                    if llevel == wl.name.lower() or llevel == wl.short_name:
                        return wl
            else:
                return cls.Auto


class EffortLevel(Enum):
    """
    This enumerated class controls the effort level used for a MIP start.
    The numeric value is identical to the CPLEX EffortLevel parameter values.

    *New in version 2.10*

    """
    Auto = 0
    CheckFeas = 1
    SolveFixed = 2
    SolveMIP = 3
    Repair = 4
    NoCheck = 5

    @classmethod
    def parse(cls, arg):
        fallback = cls.Auto
        if arg is None:
            return fallback
        elif isinstance(arg, EffortLevel):
            return arg
        else:
            for eff in EffortLevel:
                if eff.value == arg:
                    return eff
                elif is_string(arg) and arg.lower() == eff.name.lower():
                    return eff
            else:
                return fallback


# problem type conversion
_problemtype_map = {0: "LP",
                    1: "MILP",
                    3: "fixed_MILP",
                    4: "nodeLP",
                    5: "QP",
                    7: "MIQP",
                    8: "fixed_MIQP",
                    9: "node_QP",
                    10: "QCP",
                    11: "MIQCP",
                    12: "node_QCP"}


def int_probtype_to_string(probtype, fallback_probtype="unknown"):
    try:
        iprobe_type = int(probtype)
        return _problemtype_map.get(iprobe_type, fallback_probtype)
    except ValueError:
        return fallback_probtype



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

from __future__ import print_function

import sys
import copy

import six

try:  # pragma: no cover
    from itertools import zip_longest as izip_longest
except ImportError:  # pragma: no cover
    # noinspection PyUnresolvedReferences
    from itertools import izip_longest

from six import iteritems, iterkeys

from docplex.mp.compat23 import StringIO, izip
from docplex.mp.constants import CplexScope, BasisStatus, WriteLevel
from docplex.mp.utils import is_iterable, is_number, is_string, is_indexable, str_maxed, normalize_basename
from docplex.mp.utils import make_output_path2
from docplex.mp.linear import Var
from docplex.mp.error_handler import docplex_fatal

from docplex.mp.solmst import SolutionMSTPrinter
from docplex.mp.soljson import SolutionJSONPrinter

from collections import defaultdict


# noinspection PyAttributeOutsideInit
class SolveSolution(object):
    """
    The :class:`SolveSolution` class holds the result of a solve.
    """

    # a symbolic value for no objective ?
    NO_OBJECTIVE_VALUE = -1e+75

    @staticmethod
    def _is_discrete_value(v):
        return v == int(v)

    def __init__(self, model, var_value_map=None, obj=None, blended_obj_by_priority=None, name=None, solved_by=None,
                 keep_zeros=True):
        """ SolveSolution(model, var_value_map, obj, name)

        Creates a new solution object, associated to a a model.

        Args:
            model: The model to which the solution is associated. This model cannot be changed.

            obj: The value of the objective in the solution. A value of None means the objective is not defined at the
                time the solution is created, and will be set later.

            blended_obj_by_priority: For multi-objective models: the value of sub-problems' objectives (each sub-problem
                groups objectives having same priority).

            var_value_map: a Python dictionary containing associations of variables to values.

            name: a name for the solution. The default is None, in which case the solution is named after the
                model name.

        :return: A solution object.
        """
        assert model is not None
        assert solved_by is None or is_string(solved_by)
        assert obj is None or is_number(obj) or is_indexable(obj)
        assert blended_obj_by_priority is None or is_indexable(blended_obj_by_priority)

        self._model = model
        self._name = name
        self._problem_objective_expr = model.objective_expr if model.has_objective() else None
        self._objective = self.NO_OBJECTIVE_VALUE if obj is None else obj
        self._blended_objective_by_priority = [self.NO_OBJECTIVE_VALUE] if blended_obj_by_priority is None else \
            blended_obj_by_priority
        self._solved_by = solved_by
        self._var_value_map = {}

        # attributes
        self._reduced_costs = None
        self._dual_values = None
        self._slack_values = None
        self._infeasibilities = {}
        self._basis_statuses = None

        self._solve_status = None
        self._keep_zeros = keep_zeros
        self._solve_details = None

        if var_value_map is not None:
            self._store_var_value_map(var_value_map, keep_zeros=keep_zeros)

    @property
    def _checker(self):
        return self._model._checker

    @staticmethod
    def make_engine_solution(model, var_value_map, obj, blended_obj_by_priority, solved_by, solve_details,
                             job_solve_status=None):
        # INTERNAL
        # noinspection PyArgumentEqualDefault
        sol = SolveSolution(model,
                            var_value_map=None,
                            obj=obj,
                            blended_obj_by_priority=blended_obj_by_priority,
                            solved_by=solved_by,
                            keep_zeros=False)
        if solve_details is not None:
            sol._solve_details = copy.copy(solve_details)

        if model.round_solution:
            # only for models which specify round_solution
            roundfn = sol._model._round_function
            for dvar, value in iteritems(var_value_map):
                if value and dvar.is_discrete() and value != int(value):
                    var_value_map[dvar] = roundfn(value)
        # do trust engines...
        sol._var_value_map = var_value_map

        if job_solve_status is not None:
            sol._set_solve_status(job_solve_status)
        return sol

    def _get_var_by_name(self, varname):
        return self._model.get_var_by_name(varname)

    def as_mip_start(self):
        # INTERNAL
        mipstart = SolveSolution(self.model, name=self.name,
                                 var_value_map=None,
                                 obj=self.objective_value,
                                 solved_by=self.solved_by,
                                 keep_zeros=self._keep_zeros)
        # copy value dict
        mipstart._var_value_map = self._var_value_map.copy()
        if self.solved_by and not self._keep_zeros:
            # completion: add all discrete, non-generated:
            for dvi in self.model.generate_user_variables():
                if dvi.is_discrete() and dvi not in self._var_value_map:
                    mipstart._set_var_value_internal(dvi, 0)
        return mipstart

    def clear(self):
        """ Clears all solve result data.

        All data related to the model are left unchanged.
        """
        self._var_value_map = {}
        self._objective = self.NO_OBJECTIVE_VALUE
        self._reduced_costs = None
        self._dual_values = None
        self._slack_values = None
        self._infeasibilities = {}
        self._solve_status = None

    def is_empty(self):
        """
        Checks whether the solution is empty.

        Returns:
            Boolean: True if the solution is empty; in other words, the solution has no defined objective and no variable value.
        """
        return not self.has_objective() and not self._var_value_map

    @property
    def problem_name(self):
        return self._model.name

    @property
    def solved_by(self):
        '''
        Returns a string indicating how the solution was produced.

        - If the solution was created by a program, this field returns None.
        - If the solution originated from a local CPLEX solve, this method returns the string 'cplex_local'.
        - If the solution originated from a DOcplexcloud solve, this method returns 'cplex_cloud'.

        Returns:
            A string, or None.

        '''
        return self._solved_by

    def get_name(self):
        return self._name

    def set_name(self, solution_name):
        self._checker.typecheck_string(solution_name, accept_empty=False, accept_none=True,
                                       caller='SolveSolution.set_name(): ')
        self._name = solution_name

    @property
    def name(self):
        """ This property allows to get/set a name on the solution.

        In some cases , it might be interesting to build different solutions for the same model,
        in this case, use the name property to distinguish them.

        """
        return self._name

    @name.setter
    def name(self, sol_name):
        self.set_name(sol_name)

    def _resolve_var(self, var_key, do_raise):
        # INTERNAL: accepts either strings or variable objects
        # returns a variable or None
        if isinstance(var_key, Var):
            return var_key
        elif is_string(var_key):
            var = self._get_var_by_name(var_key)
            # var might be None here if the name is unknown
            if var is not None:
                return var
            # var is None hereafter
            elif do_raise:
                self.model.fatal("No variable with named {0}", var_key)
            else:
                self.model.warning("No variable with named {0}", var_key)
                return None

        else:  # pragma: no cover
            self.model.fatal("Expecting variable or name, got: {0!r}", var_key)

    def _typecheck_var_key_value(self, var_key, value, caller):
        # INTERNAL
        self._checker.typecheck_num(value, caller=caller)
        if not is_string(var_key) and not isinstance(var_key, Var):
            self.model.fatal("{0} expects either Var or string, got: {1!r}", caller, var_key)

    def add_var_value(self, var_key, value):
        """ Adds a new (variable, value) pair to this solution.

        Args:
            var_key: A decision variable (:class:`docplex.mp.linear.Var`) or a variable name (string).
            value (number): The value of the variable in the solution.
        """
        self._typecheck_var_key_value(var_key, value, caller="Solution.add_var_value")
        self._set_var_key_value(var_key, value, keep_zero=self._keep_zeros)

    def __setitem__(self, var_key, value):
        # always keep zero, no warnings, no checks
        self._set_var_key_value(var_key, value, keep_zero=self._keep_zeros)

    def set_var_key_value(self, var_key, value, keep_zero):
        # INTERNAL
        self._typecheck_var_key_value(var_key, value, caller="Solution.add_var_value")
        self._set_var_key_value(var_key, value, keep_zero)

    def _set_var_key_value(self, var_key, value, keep_zero):
        # INTERNAL: no checks done.
        dvar = self._resolve_var(var_key, do_raise=False)
        if dvar is not None:
            if value or keep_zero:
                # either value is nonzero or we keep all, store.
                self._set_var_value_internal(dvar, value)
            elif self.contains(dvar):
                # value is 0 and we dont keep zeros: zap the variable, if
                del self._var_value_map[dvar]

    def _set_var_value_internal(self, var, value):
        self._var_value_map[var] = value

    def _set_var_value(self, var, value):
        # INTERNAL
        self._set_var_value_internal(var, value)

    def update(self, var_values_iterable):
        """
        Updates the solution from a dictionary. Keys can be either strings, interpreted as variable names,
        or variables; values are the new values for the variable.

        This method returns nothing, only performs a side effect on the solution object.

        :param var_values_iterable: a dictionary of keys, values.

        """
        keep_zeros = self._keep_zeros
        for k, v in iteritems(var_values_iterable):
            self._set_var_key_value(k, v, keep_zeros)

    @property
    def model(self):
        """
        This property returns the model associated with the solution.
        """
        return self._model

    @property
    def solve_details(self):
        """ This property returns the solve_details associated with the solution,if any.

        Note:
            This property returns an instance of solve details if the solution is the result
            of a solve operation. If the solution has been created by API, this property returns None

        See Also:
            :class:`docplex.mp.sdetails.SolveDetails`

        Returns:
            an instance of SolveDetails, or None.

        """
        return self._solve_details

    # @property
    # def error_handler(self):
    #     return self.__model.error_handler

    def get_objective_value(self):
        """
        Gets the objective value (or list of objectives value) as defined in the solution.
        When the objective value has not been defined, a special value `NO_SOLUTION` is returned.
        To check whether the objective has been set, use :func:`has_objective`.

        Returns:
            float or list(float): The value of the objective (or list of values for multi-objective) as defined by
            the solution.
        """
        return self._objective

    def set_objective_value(self, obj):
        """
        Sets the objective value (or list of values for multi-objective) of the solution.
        
        Args:
            obj (float or list(float)): The value of the objective (or list of values for multi-objective) in
            the solution.
        """
        self._objective = obj

    def get_blended_objective_value_by_priority(self):
        """
        Gets the blended objective value (or list of blended objectives value) by priority level as defined in
        the solution.
        When the objective value has not been defined, a special value `NO_SOLUTION` is returned.
        To check whether the objective has been set, use :func:`has_objective`.

        Returns:
            float or list(float): The value of the objective (or list of values for multi-objective) as defined by
            the solution.
        """
        return self._blended_objective_by_priority

    @property
    def blended_objective_values(self):
        return self._blended_objective_by_priority

    def has_objective(self):
        """
        Checks whether or not the objective has been set.

        Returns:
            Boolean: True if the solution defines an objective value.
        """
        return self._objective != self.NO_OBJECTIVE_VALUE

    @property
    def objective_value(self):
        """ This property is used to get the objective value of the solution.
        In case of multi-objective this property returns the value for the first objective

        When the objective value has not been defined, a special value `NO_SOLUTION` is returned.
        To check whether the objective has been set, use :func:`has_objective`.

        """
        try:
            return self._objective[0]
        except TypeError:
            return self._objective

    @objective_value.setter
    def objective_value(self, new_objvalue):
        self.set_objective_value(new_objvalue)

    @property
    def multi_objective_values(self):
        """ This property is used to get the list of objective values of the solution.
        In case of single objective this property returns the value for the objective as a singleton list

        When the objective value has not been defined, a special value `NO_SOLUTION` is returned.
        To check whether the objective has been set, use :func:`has_objective`.

        """
        self_obj = self._objective
        return self_obj if is_indexable(self_obj) else [self_obj]

    @property
    def solve_status(self):
        return self._solve_status

    def _set_solve_status(self, new_status):
        # INTERNAL
        self._solve_status = new_status

    def _store_var_value_map(self, key_value_map, keep_zeros=False):
        # INTERNAL
        for e, val in iteritems(key_value_map):
            # need to check var_keys and values
            self.set_var_key_value(var_key=e, value=val, keep_zero=keep_zeros)

    def store_infeasibilities(self, infeasibilities):
        assert isinstance(infeasibilities, dict)
        self._infeasibilities = infeasibilities

    @staticmethod
    def _resolve_attribute_index_map(attr_idx_map, obj_mapper):
        return {obj_mapper(idx): attr_val
                for idx, attr_val in iteritems(attr_idx_map)
                if attr_val and obj_mapper(idx) is not None}

    def store_reduced_costs(self, rcs, mapper):
        self._reduced_costs = self._resolve_attribute_index_map(rcs, mapper)

    def store_dual_values(self, duals, mapper):
        self._dual_values = self._resolve_attribute_index_map(duals, mapper)

    def store_slack_values(self, slacks, mapper):
        resolved_linear_slacks = self._resolve_attribute_index_map(slacks, mapper)
        self._slack_values = defaultdict(dict)
        self._slack_values[CplexScope.LINEAR_CT_SCOPE] = resolved_linear_slacks

    def iter_var_values(self):
        """Iterates over the (variable, value) pairs in the solution.

        Returns:
            iterator: A dict-style iterator which returns a two-component tuple (variable, value)
            for all variables mentioned in the solution.
        """
        return iteritems(self._var_value_map)

    iteritems = iter_var_values

    def iter_variables(self):
        """Iterates over all variables mentioned in the solution.

        Returns:
           iterator: An iterator object over all variables mentioned in the solution.
        """
        return iterkeys(self._var_value_map)

    def contains(self, dvar):
        """
        Checks whether or not a decision variable is mentioned in the solution.

        This predicate can also be used in the form `var in solution`, because the
        :func:`__contains_` method has been redefined for this purpose.

        Args:
            dvar (:class:`docplex.mp.linear.Var`): The variable to check.

        Returns:
            Boolean: True if the variable is mentioned in the solution.
        """
        return dvar in self._var_value_map

    def __contains__(self, dvar):
        return self.contains(dvar)

    def get_value(self, arg):
        """
        Gets the value of a variable or an expression in a solution.
        If the variable is not mentioned in the solution,
        the method returns 0 and does not raise an exception.
        Note that this method can also be used as :func:`solution[arg]`
        because the :func:`__getitem__` method has been overloaded.

        Args:
            arg: A decision variable (:class:`docplex.mp.linear.Var`),
                 a variable name (a string), or an expression.

        Returns:
            float: The value of the variable in the solution.
        """
        if is_string(arg):
            var = self._get_var_by_name(arg)
            if var is None:
                self.model.fatal("No variable with name: {0}", arg)
            else:
                return self._get_var_value(var)
        elif isinstance(arg, Var):
            return self._get_var_value(arg)
        else:
            try:
                v = arg._get_solution_value(self)
                return v
            except AttributeError:
                self._model.fatal("Expecting variable, variable name or expression, {0!r} was passed", arg)

    def get_var_value(self, dvar):
        self._checker.typecheck_var(dvar)
        return self._get_var_value(dvar)

    def _get_var_value(self, dvar):
        # INTERNAL
        return self._var_value_map.get(dvar, 0)

    def get_value_list(self, dvars):
        """
        Gets the value of a sequence of variables in a solution.
        If a variable is not mentioned in the solution,
        the method assumes 0 and does not raise an exception.

        Args:
            dvars: an ordered sequence of decision variables.

        Returns:
            list: A list of float values, in the same order as the variable sequence.

        """
        checker = self._checker
        checker.check_ordered_sequence(arg=dvars,
                                       caller='SolveSolution.get_values() expects ordered sequence of variables')
        dvar_seq = checker.typecheck_var_seq(dvars)
        return self._get_values(dvar_seq)

    def get_values(self, var_seq):
        """ Same as get_value_list
        """
        return self.get_value_list(var_seq)

    def _get_values(self, dvars):
        # internal: no checks are done.
        self_value_map = self._var_value_map
        return [self_value_map.get(dv, 0) for dv in dvars]

    def get_all_values(self):
        # internal: no checks are done.
        self_value_map = self._var_value_map
        m = self._model
        return [self_value_map.get(dv, 0) for dv in m.iter_variables()]

    def get_value_dict(self, var_dict, keep_zeros=True, precision=0):
        """ Converts a dictionary of variables to a dictionary of solutions

        Assuming `var_dict` is a dictionary of variables
        (for example, as returned by `Model.integer_var_dict()`,
        returns a dictionary with the same keys and as values the solution values of the
        variables.

        :param var_dict: a dictionary of decision variables.
        :param keep_zeros: an optional flag to keep zero values (default is True)
        :param precision: an optional precision, used when filtering out zero values. The default is 1e-6.
            When keep_zeros is False, all values smaller than this value are left out.

        :return: A dictionary from variable keys to solution values (floats).
        """
        # assume var_dict is a key-> variable dictionary
        if keep_zeros:
            return {k: self._get_var_value(v) for k, v in iteritems(var_dict)}
        else:
            value_dict = {}
            for key, dvar in iteritems(var_dict):
                dvar_value = self._get_var_value(dvar)
                if (precision and abs(dvar_value) >= precision) or dvar_value:
                    value_dict[key] = dvar_value
            return value_dict

    def get_value_df(self, var_dict, value_column_name=None, key_column_names=None):
        """ Returns values of a dicitonary of variables, as a pandas dataframe.

        If pandas is not present, returns a dicitonary of columns.

        :param var_dict: the dicitonary of variables, as created by Model.xx_var_dict
        :param value_column_name: an optional string to name the value column. Default is 'value'
        :param key_column_names: an optional list of strings to name th ekeys of the dicitonary.
            If not present, keys are named 'k1', 'k2', ...

        :return: a pandas DataFrame, if pandas is present.
        """
        keys = list(six.iterkeys(var_dict))
        values = self.get_values(six.itervalues(var_dict))
        if isinstance(keys[0], tuple):
            keys = list(zip(*keys))
            knames = None
            if key_column_names:
                if len(key_column_names) == len(keys):
                    knames = key_column_names
            if not knames:
                knames = ['key_%d' % k for k in range(1, len(keys)+1)]
            kd = {kn: ks for kn, ks in zip(knames, keys)}
        else:
            kn = key_column_names or 'key'
            kd = {kn: keys}
        value_col_name = value_column_name or 'value'
        kd[value_col_name] = values
        try:
            import pandas as pd
            return pd.DataFrame(kd)
        except ImportError:
            self.model.warning("pandas module not found, returning a dict instead of DataFrame")
            return kd

    # def __len__(self):
    #     return len(self.__var_value_map)

    @property
    def number_of_var_values(self):
        """ This property returns the number of variable values stored in this solution.

        """
        return len(self._var_value_map)

    def __getitem__(self, arg):
        return self.get_value(arg)

    def get_status(self, ct):
        """ Returns the status of a linear constraint in the solution.

        Returns 1 if the constraint is satisfied, else returns 0. This is particularly useful when using
        the status variable of constraints.

        :param ct: A linear constraint
        :return: a number (1 or 0)
        """
        self._checker.typecheck_linear_constraint(ct)
        return self._get_status(ct)

    def _get_status(self, ct):
        # INTERNAL
        ct_status_var = ct._get_status_var()
        if ct_status_var:
            return self._var_value_map.get(ct_status_var, 0)
        elif ct.is_added():
            # a posted constraint is true if there is a solution
            return 1
        else:
            return 1 if ct.is_satisfied(self) else 0

    def find_unsatisfied_constraints(self, m, tolerance=1e-6):
        unsats = []
        for ct in m.iter_constraints():
            if not ct.is_satisfied(self, tolerance):
                unsats.append(ct)
        return unsats

    def restore(self, mdl, tolerance=1e-6, restore_all=False):
        # restores the solution in its model, adding ranges.
        lfactory = mdl._lfactory
        restore_ranges = []
        for dvar, val in self.iter_var_values():
            if not dvar.is_generated() or restore_all:
                rlb = max(dvar.lb, val - tolerance)
                rub = min(dvar.ub, val + tolerance)
                restore_ranges.append(lfactory.new_range_constraint(rlb, dvar, rub))
        mdl.info("restored {0} variable values using range constraints".format(len(restore_ranges)))
        return mdl.add(restore_ranges)

    def find_invalid_domain_variables(self, m, tolerance=1e-6):
        invalid_domain_vars = []
        for dv in m.iter_variables():
            dvv = self.get_var_value(dv)
            if not dv.accepts_value(dvv, tolerance=tolerance):
                invalid_domain_vars.append(dv)
        return invalid_domain_vars

    def is_valid_solution(self, tolerance=1e-6, silent=True):
        """ Returns True if the solution is valid.

        This method checks that solution values for variables are acceptable for their types
        and bounds. It also checks that all constraints are satisfied with these values.

        :param tolerance: a float number used to check constraint satisfaction. default is 1e-6.
        :param silent: optional flag. If False, prints which avriable or constraint causes the solution to be invalid.
            default is True (nothing is printed)

        :return: True if the solution is valid, within the tolerance value.

        *New in version 2.13*
        """
        m = self.model
        invalid_domain_vars = self.find_invalid_domain_variables(m, tolerance)
        if not silent and invalid_domain_vars:
            m.warning("invalid domain vars: {0}".format(len(invalid_domain_vars)))
            for v, ivd in enumerate(invalid_domain_vars, start=1):
                dvv = self.get_var_value(ivd)
                m.warning("{0} - invalid value {1} for variable {2!s}".format(v, dvv, ivd))

        unsat_cts = self.find_unsatisfied_constraints(m, tolerance)
        if not silent and unsat_cts:
            m.info("unsatisfied constraints[{0}]".format(len(unsat_cts)))
            for u, uct in enumerate(unsat_cts, start=1):
                m.warning("{0} - unsatisified constraint: {1!s}".format(u, uct))
        return not (invalid_domain_vars or unsat_cts)

    is_feasible_solution = is_valid_solution

    def equals_solution(self, other, check_models=False, obj_precision=1e-3, var_precision=1e-6):
        from itertools import dropwhile
        if check_models and (self.model != other.model):
            return False

        if is_iterable(self.objective_value) and is_iterable(other.objective_value):
            if len(self.objective_value) == len(other.objective_value):
                for self_obj_val, other_obj_val in zip(self.objective_value, other.objective_value):
                    if abs(self_obj_val - other_obj_val) >= obj_precision:
                        return False
            else:  # Different number of objectives
                return False
        elif not is_iterable(self.objective_value) and not is_iterable(other.objective_value):
            if abs(self.objective_value - other.objective_value) >= obj_precision:
                return False
        else:  # One solution is for multi-objective, and not the other
            return False

        # noinspection PyPep8
        this_triplets  = [(dv.index, dv.name, svalue) for dv, svalue in dropwhile(lambda dvv: not dvv[1],
                                                                                  self.iter_var_values())]
        other_triplets = [(dv.index, dv.name, svalue) for dv, svalue in dropwhile(lambda dvv: not dvv[1],
                                                                                  other.iter_var_values())]
        # noinspection PyArgumentList
        for this_triple, other_triple in izip(this_triplets, other_triplets):
            this_index, this_name, this_val = this_triple
            other_index, other_name, other_val = other_triple
            if other_index != this_index or this_name != other_name or \
                    abs(this_val - other_val) >= var_precision:
                return False
        else:
            return True

    def ensure_reduced_costs(self, model, engine):
        if self._reduced_costs is None:
            self._reduced_costs = engine.get_all_reduced_costs(model)

    def ensure_dual_values(self, model, engine):
        if self._dual_values is None:
            self._dual_values = engine.get_all_dual_values(model)

    def ensure_slack_values(self, model, engine):
        if self._slack_values is None:
            self._slack_values = engine.get_all_slack_values(model)

    def ensure_basis_statuses(self, model, engine):
        if self._basis_statuses is None:
            #  returns a tuple of two lists
            self._basis_statuses = engine.get_basis(model)

    def has_basis(self):
        m = self.model
        self.ensure_basis_statuses(m, m.get_engine())
        return self._has_basis()

    def _has_basis(self):
        try:
            return len(self._basis_statuses[0]) > 0
        except TypeError:
            return False

    def get_reduced_costs(self, dvars):
        m = self.model
        self.ensure_reduced_costs(m, m.get_engine())
        rcs = self._reduced_costs
        assert rcs is not None
        return [rcs.get(dv, 0) for dv in dvars]

    def get_dual_values(self, lcts):
        duals = self._dual_values
        assert duals is not None
        return [duals.get(lc, 0) for lc in lcts]

    def get_slacks(self, cts):
        all_slacks = self._slack_values
        assert all_slacks is not None
        # first get cplex_scope, then fetch the slack: two indirections
        return [all_slacks[ct.cplex_scope].get(ct, 0) for ct in cts]

    def get_var_basis_statuses(self, dvars):
        assert self._basis_statuses is not None
        all_var_basis_statuses = self._basis_statuses[0]
        return [BasisStatus.parse(all_var_basis_statuses.get(dv, -1)) for dv in dvars]

    def get_linearct_basis_statuses(self, linear_cts):
        assert self._basis_statuses is not None
        all_linearct_basis_statuses = self._basis_statuses[1]
        return [BasisStatus.parse(all_linearct_basis_statuses.get(lct, -1)) for lct in linear_cts]

    def get_infeasibility(self, ct):
        return self._infeasibilities.get(ct, 0)

    def display_attributes(self):
        pass

    def display(self,
                print_zeros=True,
                header_fmt="solution for: {0:s}",
                objective_fmt="{0}: {1:.{prec}f}",
                value_fmt="{varname:s} = {value:.{prec}f}",
                iter_vars=None,
                **kwargs):
        print_generated = kwargs.get("print_generated", False)
        problem_name = self.problem_name
        if header_fmt and problem_name:
            print(header_fmt.format(problem_name))
        if self._problem_objective_expr is not None and objective_fmt and self.has_objective():
            obj_prec = self.model.objective_expr.float_precision
            obj_name = self._problem_objective_name()
            print(objective_fmt.format(obj_name, self._objective, prec=obj_prec))
        if iter_vars is None:
            iter_vars = self.iter_variables()
        print_counter = 0
        for dvar in iter_vars:
            if print_generated or not dvar.is_generated():
                var_value = self._get_var_value(dvar)
                if print_zeros or var_value:
                    print_counter += 1
                    varname = dvar.to_string()
                    if type(value_fmt) != type(varname):
                        # infamous mix of str and unicode. Should happen only
                        # in py2. Let's convert things
                        if isinstance(value_fmt, str):
                            # noinspection PyUnresolvedReferences
                            value_fmt = value_fmt.decode('utf-8')
                        else:
                            value_fmt = value_fmt.encode('utf-8')
                    output = value_fmt.format(varname=varname,
                                              value=var_value,
                                              prec=dvar.float_precision,
                                              counter=print_counter)
                    try:
                        print(output)
                    except UnicodeEncodeError:
                        encoding = 'ascii'
                        if hasattr(sys.stdout, 'encoding') and sys.stdout.encoding:
                            encoding = sys.stdout.encoding
                        print(output.encode(encoding,
                                            errors='backslashreplace'))

    def to_string(self, print_zeros=True):
        oss = StringIO()
        self.to_stringio(oss, print_zeros=print_zeros)
        return oss.getvalue()

    def _problem_objective_name(self, default_obj_name="objective"):
        # INTERNAL
        # returns the string used for displaying the objective
        # if the problem has an objective with a name, use it
        # else return the default (typically "objective"
        self_objective_expr = self._problem_objective_expr
        if self_objective_expr is not None and self_objective_expr.has_name():
            return self_objective_expr.name
        else:
            return default_obj_name

    def to_stringio(self, oss, print_zeros=True):
        problem_name = self.problem_name
        if problem_name:
            oss.write("solution for: %s\n" % problem_name)
        if self._problem_objective_expr is not None and self.has_objective():
            obj_name = self._problem_objective_name()
            oss.write("%s: %g\n" % (obj_name, self._objective))

        value_fmt = "{var:s}={value:.{prec}f}"
        for dvar, val in self.iter_var_values():
            if not dvar.is_generated():
                var_value = self._get_var_value(dvar)
                if print_zeros or var_value != 0:
                    oss.write(value_fmt.format(var=str(dvar), value=var_value, prec=dvar.float_precision))
                    oss.write("\n")

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        if self.has_objective():
            s_obj = "obj={0:g}".format(self.objective_value)
        else:
            s_obj = "obj=N/A"
        s_values = ",".join(["{0!s}:{1:g}".format(var, val) for var, val in iteritems(self._var_value_map)])
        r = "docplex.mp.solution.SolveSolution({0},values={{{1}}})".format(s_obj, s_values)
        return str_maxed(r, maxlen=72)

    def __iter__(self):
        # INTERNAL: this is necessary to prevent solution from being an iterable.
        # as it follows getitem protocol, it can mistakenly be interpreted as an iterable
        raise TypeError

    def __as_df__(self, name_key='name', value_key='value'):
        return self.as_df(name_key, value_key)

    def as_df(self, name_key='name', value_key='value'):
        """ Converts the solution to a pandas dataframe with two columns: variable name and values

        :param name_key: column name for variable names. Default is 'name'
        :param value_key: cilumn name for values., Default is 'value'.

        :return: a pandas dataframe, if pandas is present.

        *New in version 2.15*
        """
        assert name_key
        assert value_key
        assert name_key != value_key
        try:
            import pandas as pd
        except ImportError:
            raise ImportError('Cannot convert solution to pandas.DataFrame if pandas is not available')

        names = []
        values = []
        for dv, dvv in self.iter_var_values():
            names.append(dv.to_string())
            values.append(dvv)
        name_value_dict = {name_key: names, value_key: values}
        return pd.DataFrame(name_value_dict)

    def print_mst(self, out=sys.stdout, **kwargs):
        """ Writes the solution in MST format in an output stream (default is sys.out)
        """
        self.export(out, format='mst', **kwargs)

    def _export_as_string(self, fmt, **kwargs):
        oss = StringIO()
        printer = self.get_printer(fmt)
        printer.print_to_stream(self, oss, **kwargs)
        return oss.getvalue()

    def export_as_mst_string(self, write_level=WriteLevel.Auto, **kwargs):
        kwargs['write_level'] = WriteLevel.parse(write_level)
        return self._export_as_string(fmt='mst', **kwargs)

    def export_as_mst(self, path=None, basename=None, write_level=WriteLevel.Auto, **kwargs):
        """ Exports a solution to a file in CPLEX mst format.

        Args:
            basename: Controls the basename with which the solution is printed.
                Accepts None, a plain string, or a string format.
                If None, the model's name is used.
                If passed a plain string, the string is used in place of the model's name.
                If passed a string format (either with %s or {0}), this format is used to format the
                model name to produce the basename of the written file.
            path: A path to write the file, expects a string path or None.
                Can be a directory, in which case the basename
                that was computed with the basename argument is appended to the directory to produce
                the file.
                If given a full path, the path is directly used to write the file, and
                the basename argument is not used.
                If passed None, the output directory will be ``tempfile.gettempdir()``.
            write_level: an enumerated value which controls which variables are printed.
                The default is WriteLevel.Auto, which prints the values of all discrete variables.
                This parameter also accepts the number values of the corresponding CPLEX parameters
                (1 for AllVars, 2 for DiscreteVars, 3 for NonZeroVars, 4 for NonZeroDiscreteVars)

        Returns:
            The full path of the file, when successful, else None

        Examples:
            Assuming the solution has the name "prob":

            ``sol.export_as_mst()`` will write file prob.mst in a temporary directory.

            ``sol.export_as_mst(write_level=WriteLevel.ALlvars)`` will write file prob.mst in a temporary directory,
                and will print all variables in the problem.

            ``sol.export_as_mst(path="c:/temp/myprob1.mst")`` will write file "c:/temp/myprob1.mst".

            ``sol.export_as_mst(basename="my_%s_mipstart", path ="z:/home/")`` will write "z:/home/my_prob_mipstart.mst".

        See Also:
            :class:`docplex.mp.constants.WriteLevel`
        """
        sol_basename = normalize_basename(self.problem_name, force_lowercase=True)
        mst_path = make_output_path2(actual_name=sol_basename,
                                     extension=SolutionMSTPrinter.mst_extension,
                                     path=path,
                                     basename_fmt=basename)
        if mst_path:
            kwargs2 = kwargs.copy()
            kwargs2['write_level'] = WriteLevel.parse(write_level)
            SolutionMSTPrinter.print_to_stream(self, mst_path, **kwargs2)
            return mst_path

    @classmethod
    def get_printer(cls, key):
        # INTERNAL
        printers = {'json': SolutionJSONPrinter,
                    'xml': SolutionMSTPrinter,
                    'mst': SolutionMSTPrinter
                    }

        printer = printers.get(key.lower())
        if not printer:
            raise ValueError("format key must be one of {}".format(printers.keys()))
        return printer

    def export(self, file_or_filename, format="json", **kwargs):
        """ Export this solution.
        
        Args:
            file_or_filename: If ``file_or_filename`` is a string, this argument contains the filename to
                write to. If this is a file object, this argument contains the file object to write to.
            format: The format of the solution. The format can be:
                - "json"
                - "mst": the MST cplex format for MIP starts
                - "xml": same as MST
            kwargs: The kwargs passed to the actual exporter
        """

        printer = self.get_printer(format)

        if isinstance(file_or_filename, six.string_types):
            fp = open(file_or_filename, "w")
            close_fp = True
        else:
            fp = file_or_filename
            close_fp = False
        try:
            printer.print_to_stream(self, fp, **kwargs)
        finally:
            if close_fp:
                fp.close()

    def export_as_json_string(self, **kwargs):
        """ Returns the solution as a string in JSON format.

        :return: a string.

        *New in version 2.10*
        """
        return self._export_as_string(fmt='json', **kwargs)

    def check_as_mip_start(self, strong_check=False):
        """Checks that this solution is a valid MIP start.

        To be valid, it must have:

            * at least one discrete variable (integer or binary), and
            * the values for decision variables should be consistent with the type.

        Returns:
            Boolean: True if this solution is a valid MIP start.
        """
        count_values = 0
        count_errors = 0
        m = self.model
        for dv, dvv in self.iter_var_values():
            if dv.is_discrete() and not dv.is_generated():
                count_values += 1
                if not dv.accepts_value(dvv):  # pragma: no cover
                    count_errors += 1
                    m.warning("Solution value {1} is outside the domain of variable {0!r}: {1}, type: {2!s}",
                              dv, dvv, dv.vartype.short_name)
        if count_values == 0:
            docplex_fatal("MIP start contains no discrete variable")  # pragma: no cover
        return not count_errors if strong_check else True

    def as_dict(self, keep_zeros=False):
        var_value_dict = {}
        # INTERNAL: return a dictionary of variable: variable_value
        for dvar, dval in self.iter_var_values():
            if keep_zeros or dval:
                var_value_dict[dvar] = dval
        return var_value_dict

    def as_name_dict(self, keep_zeros=False):
        # INTERNAL: return a dictionary of variable_name: variable_value
        var_value_dict = {}
        for dvar, dval in self.iter_var_values():
            dvar_name = dvar.get_name()
            if dvar_name and (keep_zeros or dval):
                var_value_dict[dvar_name] = dval
        return var_value_dict

    def kpi_value_by_name(self, name, match_case=False):
        ''' Returns the solution value of a KPI from its name.

        Args:
            name (string): The string to be matched.

            match_case (boolean): If True, looks for a case-exact match, else
               ignores case. Default is False.

        Returns:
            The value of the KPI, evaluated in the solution.

        Note:
            This method raises an error when the string does not match any KPI in the model.

        See:
            :func: `docplex.mp.model.kpi_by_name`
        '''
        kpi = self.model.kpi_by_name(name, try_match=True, match_case=match_case)
        return kpi._get_solution_value(self)



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2019
# --------------------------------------------------------------------------

# gendoc: ignore
# This file is generated !
# See script tools/gen_version.py
docplex_version_major = 2
docplex_version_minor = 15
docplex_version_micro = 194
docplex_version_string = '2.15.194'

latest_cplex_major = 12
latest_cplex_minor = 10



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

from docplex.mp.compat23 import StringIO
from docplex.mp.constants import ComparisonType
from collections import Counter

from docplex.mp.vartype import *


class ModelStatistics(object):
    """ModelStatistics()

    This class gathers statistics from the model.

    Instances of this class are returned by the method :func:`docplex.mp.model.Model.get_statistics`.

    The class contains counters on the various types of variables and constraints
    in the model.

    """

    def __init__(self):
        self._number_of_binary_variables = 0
        self._number_of_integer_variables = 0
        self._number_of_continuous_variables = 0
        self._number_of_semicontinuous_variables = 0
        self._number_of_semiinteger_variables = 0
        self._number_of_le_constraints = 0
        self._number_of_ge_constraints = 0
        self._number_of_eq_constraints = 0
        self._number_of_range_constraints = 0
        self._number_of_indicator_constraints = 0
        self._number_of_equivalence_constraints = 0
        self._number_of_quadratic_constraints = 0

    def as_tuple(self):
        return (self._number_of_binary_variables,
                self._number_of_integer_variables,
                self._number_of_continuous_variables,
                self._number_of_semicontinuous_variables,
                self._number_of_semiinteger_variables,
                self._number_of_le_constraints,
                self._number_of_ge_constraints,
                self._number_of_eq_constraints,
                self._number_of_range_constraints,
                self._number_of_indicator_constraints,
                self._number_of_equivalence_constraints,
                self._number_of_quadratic_constraints)

    def equal_stats(self, other):
        return isinstance(other, ModelStatistics) and (self.as_tuple() == other.as_tuple())

    def __eq__(self, other):
        return self.equal_stats(other)

    def __sub__(self, other):
        if not isinstance(other, ModelStatistics):
            raise TypeError
        diffstats = ModelStatistics()
        for attr in ["_number_of_le_constraints", "_number_of_ge_constraints", "_number_of_eq_constraints"]:
            setattr(diffstats, attr, getattr(self, attr) - getattr(other, attr))
        return diffstats

    @staticmethod
    def _make_new_stats(mdl):
        # INTERNAL
        stats = ModelStatistics()
        vartype_count = Counter(type(dv.vartype) for dv in mdl.iter_variables())
        stats._number_of_binary_variables = vartype_count[BinaryVarType]
        stats._number_of_integer_variables = vartype_count[IntegerVarType]
        stats._number_of_continuous_variables = vartype_count[ContinuousVarType]
        stats._number_of_semicontinuous_variables = vartype_count[SemiContinuousVarType]
        stats._number_of_semiinteger_variables = vartype_count[SemiIntegerVarType]

        linct_count = Counter(ct.sense for ct in mdl.iter_binary_constraints())
        stats._number_of_le_constraints = linct_count[ComparisonType.LE]
        stats._number_of_eq_constraints = linct_count[ComparisonType.EQ]
        stats._number_of_ge_constraints = linct_count[ComparisonType.GE]
        stats._number_of_range_constraints = mdl.number_of_range_constraints
        stats._number_of_indicator_constraints = mdl.number_of_indicator_constraints
        stats._number_of_equivalence_constraints = mdl.number_of_equivalence_constraints
        stats._number_of_quadratic_constraints = mdl.number_of_quadratic_constraints
        return stats

    @property
    def number_of_variables(self):
        """ This property returns the total number of variables in the model.

        """
        return self._number_of_binary_variables \
               + self._number_of_integer_variables \
               + self._number_of_continuous_variables \
               + self._number_of_semicontinuous_variables

    @property
    def number_of_binary_variables(self):
        """ This property returns the number of binary decision variables in the model.

        """
        return self._number_of_binary_variables

    @property
    def number_of_integer_variables(self):
        """ This property returns the number of integer decision variables in the model.

        """
        return self._number_of_integer_variables

    @property
    def number_of_continuous_variables(self):
        """ This property returns the number of continuous decision variables in the model.

        """
        return self._number_of_continuous_variables

    @property
    def number_of_semicontinuous_variables(self):
        """ This property returns the number of semicontinuous decision variables in the model.

        """
        return self._number_of_semicontinuous_variables

    @property
    def number_of_semiinteger_variables(self):
        """ This property returns the number of semi-integer decision variables in the model.

        """
        return self._number_of_semiinteger_variables

    @property
    def number_of_linear_constraints(self):
        """ This property returns the total number of linear constraints in the model.

        This number comprises all relational constraints: <=, ==, and >=
        and also range constraints.

        """
        return self._number_of_eq_constraints + \
               self._number_of_le_constraints + \
               self._number_of_ge_constraints + \
               self._number_of_range_constraints

    @property
    def number_of_le_constraints(self):
        """ This property returns the number of <= constraints

        """
        return self._number_of_le_constraints

    @property
    def number_of_eq_constraints(self):
        """ This property returns the number of == constraints

        """
        return self._number_of_eq_constraints

    @property
    def number_of_ge_constraints(self):
        """ This property returns the number of >= constraints

        """
        return self._number_of_ge_constraints

    @property
    def number_of_range_constraints(self):
        """ This property returns the number of range constraints.

        Range constraints are of the form L <= expression <= U.

        See Also:
            :class:`docplex.mp.constr.RangeConstraint`

        """
        return self._number_of_range_constraints

    @property
    def number_of_indicator_constraints(self):
        """ This property returns the number of indicator constraints.

        See Also:
            :class:`docplex.mp.constr.IndicatorConstraint`

        """
        return self._number_of_indicator_constraints

    @property
    def number_of_equivalence_constraints(self):
        """ This property returns the number of equivalence constraints.

        See Also:
            :class:`docplex.mp.constr.EquivalenceConstraint`

        """
        return self._number_of_equivalence_constraints

    @property
    def number_of_quadratic_constraints(self):
        """ This property returns the number of quadratic constraints.

        See Also:
            :class:`docplex.mp.constr.QuadraticConstraint`

        """
        return self._number_of_quadratic_constraints

    @property
    def number_of_constraints(self):
        return self.number_of_linear_constraints +\
               self.number_of_quadratic_constraints +\
               self.number_of_indicator_constraints +\
               self._number_of_equivalence_constraints

    def print_information(self):
        """ Prints model statistics in readable format.

        """
        print(' - number of variables: {0}'.format(self.number_of_variables))
        var_fmt = '   - binary={0}, integer={1}, continuous={2}'
        if 0 != self._number_of_semicontinuous_variables:
            var_fmt += ', semi-continuous={3}'
        print(var_fmt.format(self.number_of_binary_variables,
                             self.number_of_integer_variables,
                             self.number_of_continuous_variables,
                             self._number_of_semicontinuous_variables
                             ))

        print(' - number of constraints: {0}'.format(self.number_of_constraints))
        ct_fmt = '   - linear={0}'
        if 0 != self._number_of_indicator_constraints:
            ct_fmt += ', indicator={1}'
        if 0 != self._number_of_equivalence_constraints:
            ct_fmt += ', equiv={2}'
        if 0 != self._number_of_quadratic_constraints:
            ct_fmt += ', quadratic={3}'
        print(ct_fmt.format(self.number_of_linear_constraints,
                            self.number_of_indicator_constraints,
                            self.number_of_equivalence_constraints,
                            self.number_of_quadratic_constraints))

    def to_string(self):
        oss = StringIO()
        oss.write(" - number of variables: %d\n" % self.number_of_variables)
        var_fmt = '   - binary={0}, integer={1}, continuous={2}'
        if 0 != self._number_of_semicontinuous_variables:
            var_fmt += ', semi-continuous={3}'
        oss.write(var_fmt.format(self.number_of_binary_variables,
                                 self.number_of_integer_variables,
                                 self.number_of_continuous_variables,
                                 self._number_of_semicontinuous_variables
                                 ))
        oss.write('\n')
        nb_constraints = self.number_of_constraints
        oss.write(' - number of constraints: {0}\n'.format(nb_constraints))
        if nb_constraints:
            ct_fmt = '   - linear={0}'
            if 0 != self._number_of_indicator_constraints:
                ct_fmt += ', indicator={1}'
            if 0 != self._number_of_equivalence_constraints:
                ct_fmt += ', equiv={2}'
            if 0 != self._number_of_quadratic_constraints:
                ct_fmt += ', quadratic={3}'
            oss.write(ct_fmt.format(self.number_of_linear_constraints,
                                    self.number_of_indicator_constraints,
                                    self.number_of_equivalence_constraints,
                                    self.number_of_quadratic_constraints))
        return oss.getvalue()

    def __str__(self):
        return self.to_string()

    def __repr__(self):  # pragma: no cover
        return "docplex.mp.Model.ModelStatistics()"
